@@ -1,5 +1,6 @@
-/*	$OpenBSD: cert.c,v 1.62 2022/04/05 03:56:20 tb Exp $ */
+/*	$OpenBSD: cert.c,v 1.84 2022/05/31 18:51:35 tb Exp $ */
 /*
+ * Copyright (c) 2022 Theo Buehler <tb@openbsd.org>
  * Copyright (c) 2021 Job Snijders <job@openbsd.org>
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -33,13 +34,6 @@
 
 #include "extern.h"
 
-/*
- * Type of ASIdentifier (RFC 3779, 3.2.3).
- */
-#define	ASID_TYPE_ASNUM	0x00
-#define ASID_TYPE_RDI	0x01
-#define ASID_TYPE_MAX	ASID_TYPE_RDI
-
 /*
  * A parsing sequence of a file (which may just be <stdin>).
  */
@@ -65,19 +59,12 @@ extern ASN1_OBJECT	*notify_oid;	/* 1.3.6.1.5.5.7.48.13 (rpkiNotify) */
  * Returns zero on failure (IP overlap) non-zero on success.
  */
 static int
-append_ip(struct parse *p, const struct cert_ip *ip)
+append_ip(const char *fn, struct cert_ip *ips, size_t *ipsz,
+    const struct cert_ip *ip)
 {
-	struct cert	*res = p->res;
-
-	if (!ip_addr_check_overlap(ip, p->fn, p->res->ips, p->res->ipsz))
+	if (!ip_addr_check_overlap(ip, fn, ips, *ipsz))
 		return 0;
-	if (res->ipsz >= MAX_IP_SIZE)
-		return 0;
-	res->ips = reallocarray(res->ips, res->ipsz + 1,
-	    sizeof(struct cert_ip));
-	if (res->ips == NULL)
-		err(1, NULL);
-	res->ips[res->ipsz++] = *ip;
+	ips[(*ipsz)++] = *ip;
 	return 1;
 }
 
@@ -87,391 +74,59 @@ append_ip(struct parse *p, const struct cert_ip *ip)
  * as defined by RFC 3779 section 3.3.
  */
 static int
-append_as(struct parse *p, const struct cert_as *as)
+append_as(const char *fn, struct cert_as *ases, size_t *asz,
+    const struct cert_as *as)
 {
-
-	if (!as_check_overlap(as, p->fn, p->res->as, p->res->asz))
+	if (!as_check_overlap(as, fn, ases, *asz))
 		return 0;
-	if (p->res->asz >= MAX_AS_SIZE)
-		return 0;
-	p->res->as = reallocarray(p->res->as, p->res->asz + 1,
-	    sizeof(struct cert_as));
-	if (p->res->as == NULL)
-		err(1, NULL);
-	p->res->as[p->res->asz++] = *as;
+	ases[(*asz)++] = *as;
 	return 1;
 }
 
 /*
- * Construct a RFC 3779 2.2.3.8 range by its bit string.
+ * Parse a range of AS identifiers as in 3.2.3.8.
  * Returns zero on failure, non-zero on success.
  */
-static int
-sbgp_addr(struct parse *p,
-	struct cert_ip *ip, const ASN1_BIT_STRING *bs)
-{
-
-	if (!ip_addr_parse(bs, ip->afi, p->fn, &ip->ip)) {
-		warnx("%s: RFC 3779 section 2.2.3.8: IPAddress: "
-		    "invalid IP address", p->fn);
-		return 0;
-	}
-	if (!ip_cert_compose_ranges(ip)) {
-		warnx("%s: RFC 3779 section 2.2.3.8: IPAddress: "
-		    "IP address range reversed", p->fn);
-		return 0;
-	}
-	return append_ip(p, ip);
-}
-
-/*
- * Parse the SIA notify URL, 4.8.8.1.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_sia_resource_notify(struct parse *p, const char *d, size_t dsz)
-{
-	if (p->res->notify != NULL) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "Notify location already specified", p->fn);
-		return 0;
-	}
-
-	/* Make sure it's a https:// address. */
-	if (!valid_uri(d, dsz, "https://")) {
-		warnx("%s: RFC 8182 section 3.2: bad Notify URI", p->fn);
-		return 0;
-	}
-
-	if ((p->res->notify = strndup(d, dsz)) == NULL)
-		err(1, NULL);
-
-	return 1;
-}
-
-/*
- * Parse the SIA manifest, 4.8.8.1.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_sia_resource_mft(struct parse *p, const char *d, size_t dsz)
-{
-	if (p->res->mft != NULL) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "MFT location already specified", p->fn);
-		return 0;
-	}
-
-	/* Make sure it's an MFT rsync address. */
-	if (!valid_uri(d, dsz, "rsync://")) {
-		warnx("%s: RFC 6487 section 4.8.8: bad MFT location", p->fn);
-		return 0;
-	}
-
-	if (dsz < 4 || strcasecmp(d + dsz - 4, ".mft") != 0) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "not an MFT file", p->fn);
-		return 0;
-	}
-
-	if ((p->res->mft = strndup(d, dsz)) == NULL)
-		err(1, NULL);
-
-	return 1;
-}
-
-/*
- * Parse the SIA manifest, 4.8.8.1.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_sia_resource_carepo(struct parse *p, const char *d, size_t dsz)
-{
-	if (p->res->repo != NULL) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "CA repository already specified", p->fn);
-		return 0;
-	}
-
-	/* Make sure it's an rsync:// address. */
-	if (!valid_uri(d, dsz, "rsync://")) {
-		warnx("%s: RFC 6487 section 4.8.8: bad CA repository URI",
-		    p->fn);
-		return 0;
-	}
-
-	if ((p->res->repo = strndup(d, dsz)) == NULL)
-		err(1, NULL);
-
-	return 1;
-}
-
-/*
- * Parse the SIA entries, 4.8.8.1.
- * There may be multiple different resources at this location, so throw
- * out all but the matching resource type. Currently only two entries
- * are of interest: rpkiManifest and rpkiNotify.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_sia_resource_entry(struct parse *p,
-	const unsigned char *d, size_t dsz)
-{
-	ASN1_SEQUENCE_ANY	*seq;
-	ASN1_OBJECT		*oid;
-	const ASN1_TYPE		*t;
-	int			 rc = 0, ptag;
-	long			 plen;
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-	if (sk_ASN1_TYPE_num(seq) != 2) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "want 2 elements, have %d",
-		    p->fn, sk_ASN1_TYPE_num(seq));
-		goto out;
-	}
-
-	/* Composed of an OID and its continuation. */
-
-	t = sk_ASN1_TYPE_value(seq, 0);
-	if (t->type != V_ASN1_OBJECT) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "want ASN.1 object, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	oid = t->value.object;
-
-	t = sk_ASN1_TYPE_value(seq, 1);
-	if (t->type != V_ASN1_OTHER) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "want ASN.1 external, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-
-	/* FIXME: there must be a way to do this without ASN1_frame. */
-
-	d = t->value.asn1_string->data;
-	dsz = t->value.asn1_string->length;
-	if (!ASN1_frame(p->fn, dsz, &d, &plen, &ptag))
-		goto out;
-
-	if (OBJ_cmp(oid, carepo_oid) == 0)
-		rc = sbgp_sia_resource_carepo(p, d, plen);
-	else if (OBJ_cmp(oid, manifest_oid) == 0)
-		rc = sbgp_sia_resource_mft(p, d, plen);
-	else if (OBJ_cmp(oid, notify_oid) == 0)
-		rc = sbgp_sia_resource_notify(p, d, plen);
-	else
-		rc = 1;	/* silently ignore */
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
-}
-
-/*
- * Multiple locations as defined in RFC 6487, 4.8.8.1.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_sia_resource(struct parse *p, const unsigned char *d, size_t dsz)
-{
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*t;
-	int			 rc = 0, i;
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-
-	for (i = 0; i < sk_ASN1_TYPE_num(seq); i++) {
-		t = sk_ASN1_TYPE_value(seq, i);
-		if (t->type != V_ASN1_SEQUENCE) {
-			warnx("%s: RFC 6487 section 4.8.8: SIA: "
-			    "want ASN.1 sequence, have %s (NID %d)",
-			    p->fn, ASN1_tag2str(t->type), t->type);
-			goto out;
-		}
-		d = t->value.asn1_string->data;
-		dsz = t->value.asn1_string->length;
-		if (!sbgp_sia_resource_entry(p, d, dsz))
-			goto out;
-	}
-
-	if (p->res->mft == NULL || p->res->repo == NULL) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA missing caRepository "
-		    "or rpkiManifest", p->fn);
-		goto out;
-	}
-
-	if (strstr(p->res->mft, p->res->repo) != p->res->mft) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "conflicting URIs for caRepository and rpkiManifest",
-		    p->fn);
-		goto out;
-	}
-	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
-}
-
-/*
- * Parse "Subject Information Access" extension, RFC 6487 4.8.8.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_sia(struct parse *p, X509_EXTENSION *ext)
-{
-	unsigned char		*sv = NULL;
-	const unsigned char	*d;
-	ASN1_SEQUENCE_ANY	*seq = NULL;
-	const ASN1_TYPE		*t;
-	int			 dsz, rc = 0;
-
-	if (X509_EXTENSION_get_critical(ext)) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "extension not non-critical", p->fn);
-		goto out;
-	}
-
-	if ((dsz = i2d_X509_EXTENSION(ext, &sv)) < 0) {
-		cryptowarnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "failed extension parse", p->fn);
-		goto out;
-	}
-	d = sv;
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-	if (sk_ASN1_TYPE_num(seq) != 2) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "want 2 elements, have %d", p->fn,
-		    sk_ASN1_TYPE_num(seq));
-		goto out;
-	}
-
-	t = sk_ASN1_TYPE_value(seq, 0);
-	if (t->type != V_ASN1_OBJECT) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "want ASN.1 object, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	if (OBJ_obj2nid(t->value.object) != NID_sinfo_access) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "incorrect OID, have %s (NID %d)", p->fn,
-		    ASN1_tag2str(OBJ_obj2nid(t->value.object)),
-		    OBJ_obj2nid(t->value.object));
-		goto out;
-	}
-
-	t = sk_ASN1_TYPE_value(seq, 1);
-	if (t->type != V_ASN1_OCTET_STRING) {
-		warnx("%s: RFC 6487 section 4.8.8: SIA: "
-		    "want ASN.1 octet string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-
-	d = t->value.octet_string->data;
-	dsz = t->value.octet_string->length;
-	if (!sbgp_sia_resource(p, d, dsz))
-		goto out;
-
-	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	free(sv);
-	return rc;
-}
-
-/*
- * Parse a range of addresses as in 3.2.3.8.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_asrange(struct parse *p, const unsigned char *d, size_t dsz)
+int
+sbgp_as_range(const char *fn, struct cert_as *ases, size_t *asz,
+    const ASRange *range)
 {
 	struct cert_as		 as;
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*t;
-	int			 rc = 0;
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 3.2.3.8: ASRange: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-	if (sk_ASN1_TYPE_num(seq) != 2) {
-		warnx("%s: RFC 3779 section 3.2.3.8: ASRange: "
-		    "want 2 elements, have %d", p->fn,
-		    sk_ASN1_TYPE_num(seq));
-		goto out;
-	}
 
 	memset(&as, 0, sizeof(struct cert_as));
 	as.type = CERT_AS_RANGE;
 
-	t = sk_ASN1_TYPE_value(seq, 0);
-	if (t->type != V_ASN1_INTEGER) {
-		warnx("%s: RFC 3779 section 3.2.3.8: ASRange: "
-		    "want ASN.1 integer, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	if (!as_id_parse(t->value.integer, &as.range.min)) {
+	if (!as_id_parse(range->min, &as.range.min)) {
 		warnx("%s: RFC 3779 section 3.2.3.8 (via RFC 1930): "
-		    "malformed AS identifier", p->fn);
+		    "malformed AS identifier", fn);
 		return 0;
 	}
 
-	t = sk_ASN1_TYPE_value(seq, 1);
-	if (t->type != V_ASN1_INTEGER) {
-		warnx("%s: RFC 3779 section 3.2.3.8: ASRange: "
-		    "want ASN.1 integer, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	if (!as_id_parse(t->value.integer, &as.range.max)) {
+	if (!as_id_parse(range->max, &as.range.max)) {
 		warnx("%s: RFC 3779 section 3.2.3.8 (via RFC 1930): "
-		    "malformed AS identifier", p->fn);
+		    "malformed AS identifier", fn);
 		return 0;
 	}
 
 	if (as.range.max == as.range.min) {
 		warnx("%s: RFC 3379 section 3.2.3.8: ASRange: "
-		    "range is singular", p->fn);
-		goto out;
+		    "range is singular", fn);
+		return 0;
 	} else if (as.range.max < as.range.min) {
 		warnx("%s: RFC 3379 section 3.2.3.8: ASRange: "
-		    "range is out of order", p->fn);
-		goto out;
+		    "range is out of order", fn);
+		return 0;
 	}
 
-	if (!append_as(p, &as))
-		goto out;
-	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
+	return append_as(fn, ases, asz, &as);
 }
 
 /*
  * Parse an entire 3.2.3.10 integer type.
  */
-static int
-sbgp_asid(struct parse *p, const ASN1_INTEGER *i)
+int
+sbgp_as_id(const char *fn, struct cert_as *ases, size_t *asz,
+    const ASN1_INTEGER *i)
 {
 	struct cert_as	 as;
 
@@ -480,97 +135,27 @@ sbgp_asid(struct parse *p, const ASN1_INTEGER *i)
 
 	if (!as_id_parse(i, &as.id)) {
 		warnx("%s: RFC 3779 section 3.2.3.10 (via RFC 1930): "
-		    "malformed AS identifier", p->fn);
+		    "malformed AS identifier", fn);
 		return 0;
 	}
 	if (as.id == 0) {
 		warnx("%s: RFC 3779 section 3.2.3.10 (via RFC 1930): "
-		    "AS identifier zero is reserved", p->fn);
+		    "AS identifier zero is reserved", fn);
 		return 0;
 	}
 
-	return append_as(p, &as);
+	return append_as(fn, ases, asz, &as);
 }
 
-/*
- * Parse one of RFC 3779 3.2.3.2.
- * Returns zero on failure, non-zero on success.
- */
 static int
-sbgp_asnum(struct parse *p, const unsigned char *d, size_t dsz)
+sbgp_as_inherit(const char *fn, struct cert_as *ases, size_t *asz)
 {
-	struct cert_as		 as;
-	ASN1_TYPE		*t, *tt;
-	ASN1_SEQUENCE_ANY	*seq = NULL;
-	int			 i, rc = 0;
-	const unsigned char	*sv = d;
+	struct cert_as as;
 
-	/* We can either be a null (inherit) or sequence. */
+	memset(&as, 0, sizeof(struct cert_as));
+	as.type = CERT_AS_INHERIT;
 
-	if ((t = d2i_ASN1_TYPE(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 3.2.3.2: ASIdentifierChoice: "
-		    "failed ASN.1 type parse", p->fn);
-		goto out;
-	}
-
-	/*
-	 * Section 3779 3.2.3.3 is to inherit with an ASN.1 NULL type,
-	 * which is the easy case.
-	 */
-
-	switch (t->type) {
-	case V_ASN1_NULL:
-		memset(&as, 0, sizeof(struct cert_as));
-		as.type = CERT_AS_INHERIT;
-		if (!append_as(p, &as))
-			goto out;
-		rc = 1;
-		goto out;
-	case V_ASN1_SEQUENCE:
-		break;
-	default:
-		warnx("%s: RFC 3779 section 3.2.3.2: ASIdentifierChoice: "
-		    "want ASN.1 sequence or null, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-
-	/* This is RFC 3779 3.2.3.4. */
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &sv, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 3.2.3.2: ASIdentifierChoice: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-
-	/* Accepts RFC 3779 3.2.3.6 or 3.2.3.7 (sequence). */
-
-	for (i = 0; i < sk_ASN1_TYPE_num(seq); i++) {
-		tt = sk_ASN1_TYPE_value(seq, i);
-		switch (tt->type) {
-		case V_ASN1_INTEGER:
-			if (!sbgp_asid(p, tt->value.integer))
-				goto out;
-			break;
-		case V_ASN1_SEQUENCE:
-			d = tt->value.asn1_string->data;
-			dsz = tt->value.asn1_string->length;
-			if (!sbgp_asrange(p, d, dsz))
-				goto out;
-			break;
-		default:
-			warnx("%s: RFC 3779 section 3.2.3.5: ASIdOrRange: "
-			    "want ASN.1 sequence or integer, have %s (NID %d)",
-			    p->fn, ASN1_tag2str(tt->type), tt->type);
-			goto out;
-		}
-	}
-
-	rc = 1;
-out:
-	ASN1_TYPE_free(t);
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
+	return append_as(fn, ases, asz, &as);
 }
 
 /*
@@ -581,12 +166,10 @@ out:
 static int
 sbgp_assysnum(struct parse *p, X509_EXTENSION *ext)
 {
-	unsigned char		*sv = NULL;
-	const unsigned char	*d;
-	ASN1_SEQUENCE_ANY	*seq = NULL, *sseq = NULL;
-	const ASN1_TYPE		*t;
-	int			 dsz, rc = 0, i, ptag;
-	long			 plen;
+	ASIdentifiers		*asidentifiers = NULL;
+	const ASIdOrRanges	*aors = NULL;
+	size_t			 asz;
+	int			 i, rc = 0;
 
 	if (!X509_EXTENSION_get_critical(ext)) {
 		cryptowarnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
@@ -594,290 +177,158 @@ sbgp_assysnum(struct parse *p, X509_EXTENSION *ext)
 		goto out;
 	}
 
-	if ((dsz = i2d_X509_EXTENSION(ext, &sv)) < 0) {
+	if ((asidentifiers = X509V3_EXT_d2i(ext)) == NULL) {
 		cryptowarnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
 		    "failed extension parse", p->fn);
 		goto out;
 	}
 
-	/* Start with RFC 3779, section 3.2 top-level. */
-
-	d = sv;
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-	if (sk_ASN1_TYPE_num(seq) != 3) {
+	if (asidentifiers->rdi != NULL) {
 		warnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
-		    "want 3 elements, have %d", p->fn,
-		    sk_ASN1_TYPE_num(seq));
+		    "should not have RDI values", p->fn);
 		goto out;
 	}
 
-	t = sk_ASN1_TYPE_value(seq, 0);
-	if (t->type != V_ASN1_OBJECT) {
+	if (asidentifiers->asnum == NULL) {
 		warnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
-		    "want ASN.1 object, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
+		    "no AS number resource set", p->fn);
 		goto out;
 	}
 
-	t = sk_ASN1_TYPE_value(seq, 1);
-	if (t->type != V_ASN1_BOOLEAN) {
-		warnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
-		    "want ASN.1 boolean, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
+	switch (asidentifiers->asnum->type) {
+	case ASIdentifierChoice_inherit:
+		asz = 1;
+		break;
+	case ASIdentifierChoice_asIdsOrRanges:
+		aors = asidentifiers->asnum->u.asIdsOrRanges;
+		asz = sk_ASIdOrRange_num(aors);
+		break;
+	default:
+		warnx("%s: RFC 3779 section 3.2.3.2: ASIdentifierChoice: "
+		    "unknown type %d", p->fn, asidentifiers->asnum->type);
 		goto out;
 	}
 
-	t = sk_ASN1_TYPE_value(seq, 2);
-	if (t->type != V_ASN1_OCTET_STRING) {
-		warnx("%s: RFC 6487 section 4.8.11: autonomousSysNum: "
-		    "want ASN.1 octet string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
+	if (asz >= MAX_AS_SIZE) {
+		warnx("%s: too many AS number entries: limit %d",
+		    p->fn, MAX_AS_SIZE);
 		goto out;
 	}
+	p->res->as = calloc(asz, sizeof(struct cert_as));
+	if (p->res->as == NULL)
+		err(1, NULL);
 
-	/* Within RFC 3779 3.2.3, check 3.2.3.1. */
-
-	d = t->value.octet_string->data;
-	dsz = t->value.octet_string->length;
-
-	if ((sseq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 3.2.3.1: ASIdentifiers: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
+	if (aors == NULL) {
+		if (!sbgp_as_inherit(p->fn, p->res->as, &p->res->asz))
+			goto out;
 	}
 
-	/* Scan through for private 3.2.3.2 classes. */
+	for (i = 0; i < sk_ASIdOrRange_num(aors); i++) {
+		const ASIdOrRange *aor;
 
-	for (i = 0; i < sk_ASN1_TYPE_num(sseq); i++) {
-		t = sk_ASN1_TYPE_value(sseq, i);
-		if (t->type != V_ASN1_OTHER) {
-			warnx("%s: RFC 3779 section 3.2.3.1: ASIdentifiers: "
-			    "want ASN.1 explicit, have %s (NID %d)", p->fn,
-			    ASN1_tag2str(t->type), t->type);
+		aor = sk_ASIdOrRange_value(aors, i);
+		switch (aor->type) {
+		case ASIdOrRange_id:
+			if (!sbgp_as_id(p->fn, p->res->as, &p->res->asz,
+			    aor->u.id))
+				goto out;
+			break;
+		case ASIdOrRange_range:
+			if (!sbgp_as_range(p->fn, p->res->as, &p->res->asz,
+			    aor->u.range))
+				goto out;
+			break;
+		default:
+			warnx("%s: RFC 3779 section 3.2.3.5: ASIdOrRange: "
+			    "unknown type %d", p->fn, aor->type);
 			goto out;
 		}
-
-		/* Use the low-level ASN1_frame. */
-
-		d = t->value.asn1_string->data;
-		dsz = t->value.asn1_string->length;
-		if (!ASN1_frame(p->fn, dsz, &d, &plen, &ptag))
-			goto out;
-
-		/* Ignore bad AS identifiers and RDI entries. */
-
-		if (ptag > ASID_TYPE_MAX) {
-			warnx("%s: RFC 3779 section 3.2.3.1: ASIdentifiers: "
-			    "unknown explicit tag 0x%02x", p->fn, ptag);
-			goto out;
-		} else if (ptag == ASID_TYPE_RDI)
-			continue;
-
-		if (!sbgp_asnum(p, d, plen))
-			goto out;
 	}
 
 	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	sk_ASN1_TYPE_pop_free(sseq, ASN1_TYPE_free);
-	free(sv);
+ out:
+	ASIdentifiers_free(asidentifiers);
 	return rc;
 }
 
+/*
+ * Construct a RFC 3779 2.2.3.8 range from its bit string.
+ * Returns zero on failure, non-zero on success.
+ */
+int
+sbgp_addr(const char *fn, struct cert_ip *ips, size_t *ipsz, enum afi afi,
+    const ASN1_BIT_STRING *bs)
+{
+	struct cert_ip	ip;
+
+	memset(&ip, 0, sizeof(struct cert_ip));
+
+	ip.afi = afi;
+	ip.type = CERT_IP_ADDR;
+
+	if (!ip_addr_parse(bs, afi, fn, &ip.ip)) {
+		warnx("%s: RFC 3779 section 2.2.3.8: IPAddress: "
+		    "invalid IP address", fn);
+		return 0;
+	}
+
+	if (!ip_cert_compose_ranges(&ip)) {
+		warnx("%s: RFC 3779 section 2.2.3.8: IPAddress: "
+		    "IP address range reversed", fn);
+		return 0;
+	}
+
+	return append_ip(fn, ips, ipsz, &ip);
+}
+
 /*
  * Parse RFC 3779 2.2.3.9 range of addresses.
  * Returns zero on failure, non-zero on success.
  */
-static int
-sbgp_addr_range(struct parse *p, struct cert_ip *ip,
-	const unsigned char *d, size_t dsz)
+int
+sbgp_addr_range(const char *fn, struct cert_ip *ips, size_t *ipsz,
+    enum afi afi, const IPAddressRange *range)
 {
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*t;
-	int			 rc = 0;
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-	if (sk_ASN1_TYPE_num(seq) != 2) {
-		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
-		    "want 2 elements, have %d", p->fn, sk_ASN1_TYPE_num(seq));
-		goto out;
-	}
-
-	t = sk_ASN1_TYPE_value(seq, 0);
-	if (t->type != V_ASN1_BIT_STRING) {
-		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
-		    "want ASN.1 bit string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	if (!ip_addr_parse(t->value.bit_string,
-	    ip->afi, p->fn, &ip->range.min)) {
-		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
-		    "invalid IP address", p->fn);
-		goto out;
-	}
-
-	t = sk_ASN1_TYPE_value(seq, 1);
-	if (t->type != V_ASN1_BIT_STRING) {
-		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
-		    "want ASN.1 bit string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	if (!ip_addr_parse(t->value.bit_string,
-	    ip->afi, p->fn, &ip->range.max)) {
-		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
-		    "invalid IP address", p->fn);
-		goto out;
-	}
-
-	if (!ip_cert_compose_ranges(ip)) {
-		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
-		    "IP address range reversed", p->fn);
-		return 0;
-	}
-
-	rc = append_ip(p, ip);
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
-}
-
-/*
- * Parse an IP address or range, RFC 3779 2.2.3.7.
- * We don't constrain this parse (as specified in section 2.2.3.6) to
- * having any kind of order.
- * Returns zero on failure, non-zero on success.
- */
-static int
-sbgp_addr_or_range(struct parse *p, struct cert_ip *ip,
-	const unsigned char *d, size_t dsz)
-{
-	struct cert_ip		 nip;
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*t;
-	int			 i, rc = 0;
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 2.2.3.7: IPAddressOrRange: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-
-	/* Either RFC 3779 2.2.3.8 or 2.2.3.9. */
-
-	for (i = 0; i < sk_ASN1_TYPE_num(seq); i++) {
-		nip = *ip;
-		t = sk_ASN1_TYPE_value(seq, i);
-		switch (t->type) {
-		case V_ASN1_BIT_STRING:
-			nip.type = CERT_IP_ADDR;
-			if (!sbgp_addr(p, &nip, t->value.bit_string))
-				goto out;
-			break;
-		case V_ASN1_SEQUENCE:
-			nip.type = CERT_IP_RANGE;
-			d = t->value.asn1_string->data;
-			dsz = t->value.asn1_string->length;
-			if (!sbgp_addr_range(p, &nip, d, dsz))
-				goto out;
-			break;
-		default:
-			warnx("%s: RFC 3779 section 2.2.3.7: IPAddressOrRange: "
-			    "want ASN.1 sequence or bit string, have %s (NID %d)",
-			    p->fn, ASN1_tag2str(t->type), t->type);
-			goto out;
-		}
-	}
-
-	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
-}
-
-/*
- * Parse a sequence of address families as in RFC 3779 sec. 2.2.3.2.
- * Ignore several stipulations of the RFC (2.2.3.3).
- * Namely, we don't require entries to be ordered in any way (type, AFI
- * or SAFI group, etc.).
- * This is because it doesn't matter for our purposes: we're going to
- * validate in the same way regardless.
- * Returns zero no failure, non-zero on success.
- */
-static int
-sbgp_ipaddrfam(struct parse *p, const unsigned char *d, size_t dsz)
-{
-	struct cert_ip		 ip;
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*t;
-	int			 rc = 0;
+	struct cert_ip	ip;
 
 	memset(&ip, 0, sizeof(struct cert_ip));
 
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 2.2.3.2: IPAddressFamily: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-	if (sk_ASN1_TYPE_num(seq) != 2) {
-		warnx("%s: RFC 3779 section 2.2.3.2: IPAddressFamily: "
-		    "want 2 elements, have %d",
-		    p->fn, sk_ASN1_TYPE_num(seq));
-		goto out;
+	ip.afi = afi;
+	ip.type = CERT_IP_RANGE;
+
+	if (!ip_addr_parse(range->min, afi, fn, &ip.range.min)) {
+		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
+		    "invalid IP address", fn);
+		return 0;
 	}
 
-	/* Get address family, RFC 3779, 2.2.3.3. */
-
-	t = sk_ASN1_TYPE_value(seq, 0);
-	if (t->type != V_ASN1_OCTET_STRING) {
-		warnx("%s: RFC 3779 section 2.2.3.2: addressFamily: "
-		    "want ASN.1 octet string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
+	if (!ip_addr_parse(range->max, afi, fn, &ip.range.max)) {
+		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
+		    "invalid IP address", fn);
+		return 0;
 	}
 
-	if (!ip_addr_afi_parse(p->fn, t->value.octet_string, &ip.afi)) {
-		warnx("%s: RFC 3779 section 2.2.3.2: addressFamily: "
-		    "invalid AFI", p->fn);
-		goto out;
+	if (!ip_cert_compose_ranges(&ip)) {
+		warnx("%s: RFC 3779 section 2.2.3.9: IPAddressRange: "
+		    "IP address range reversed", fn);
+		return 0;
 	}
 
-	/* Either sequence or null (inherit), RFC 3779 sec. 2.2.3.4. */
+	return append_ip(fn, ips, ipsz, &ip);
+}
 
-	t = sk_ASN1_TYPE_value(seq, 1);
-	switch (t->type) {
-	case V_ASN1_SEQUENCE:
-		d = t->value.asn1_string->data;
-		dsz = t->value.asn1_string->length;
-		if (!sbgp_addr_or_range(p, &ip, d, dsz))
-			goto out;
-		break;
-	case V_ASN1_NULL:
-		ip.type = CERT_IP_INHERIT;
-		if (!append_ip(p, &ip))
-			goto out;
-		break;
-	default:
-		warnx("%s: RFC 3779 section 2.2.3.2: IPAddressChoice: "
-		    "want ASN.1 sequence or null, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
+static int
+sbgp_addr_inherit(const char *fn, struct cert_ip *ips, size_t *ipsz,
+    enum afi afi)
+{
+	struct cert_ip	ip;
 
-	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
+	memset(&ip, 0, sizeof(struct cert_ip));
+
+	ip.afi = afi;
+	ip.type = CERT_IP_INHERIT;
+
+	return append_ip(fn, ips, ipsz, &ip);
 }
 
 /*
@@ -888,12 +339,13 @@ out:
 static int
 sbgp_ipaddrblk(struct parse *p, X509_EXTENSION *ext)
 {
-	int			 dsz, rc = 0;
-	unsigned char		*sv = NULL;
-	const unsigned char	*d;
-	ASN1_SEQUENCE_ANY	*seq = NULL, *sseq = NULL;
-	const ASN1_TYPE		*t = NULL;
-	int			 i;
+	STACK_OF(IPAddressFamily)	*addrblk = NULL;
+	const IPAddressFamily		*af;
+	const IPAddressOrRanges		*aors;
+	const IPAddressOrRange		*aor;
+	enum afi			 afi;
+	size_t				 ipsz;
+	int				 i, j, rc = 0;
 
 	if (!X509_EXTENSION_get_critical(ext)) {
 		cryptowarnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
@@ -901,81 +353,142 @@ sbgp_ipaddrblk(struct parse *p, X509_EXTENSION *ext)
 		goto out;
 	}
 
-	if ((dsz = i2d_X509_EXTENSION(ext, &sv)) < 0) {
+	if ((addrblk = X509V3_EXT_d2i(ext)) == NULL) {
 		cryptowarnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
 		    "failed extension parse", p->fn);
 		goto out;
 	}
-	d = sv;
 
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-	if (sk_ASN1_TYPE_num(seq) != 3) {
-		warnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
-		    "want 3 elements, have %d",
-		    p->fn, sk_ASN1_TYPE_num(seq));
-		goto out;
-	}
+	for (i = 0; i < sk_IPAddressFamily_num(addrblk); i++) {
+		af = sk_IPAddressFamily_value(addrblk, i);
 
-	t = sk_ASN1_TYPE_value(seq, 0);
-	if (t->type != V_ASN1_OBJECT) {
-		warnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
-		    "want ASN.1 object, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-
-	t = sk_ASN1_TYPE_value(seq, 1);
-	if (t->type != V_ASN1_BOOLEAN) {
-		warnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
-		    "want ASN.1 boolean, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-
-	t = sk_ASN1_TYPE_value(seq, 2);
-	if (t->type != V_ASN1_OCTET_STRING) {
-		warnx("%s: RFC 6487 section 4.8.10: sbgp-ipAddrBlock: "
-		    "want ASN.1 octet string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-
-	/* The blocks sequence, RFC 3779 2.2.3.1. */
-
-	d = t->value.octet_string->data;
-	dsz = t->value.octet_string->length;
-
-	if ((sseq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 3779 section 2.2.3.1: IPAddrBlocks: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-
-	/* Each sequence element contains RFC 3779 sec. 2.2.3.2. */
-
-	for (i = 0; i < sk_ASN1_TYPE_num(sseq); i++) {
-		t = sk_ASN1_TYPE_value(sseq, i);
-		if (t->type != V_ASN1_SEQUENCE) {
-			warnx("%s: RFC 3779 section 2.2.3.2: IPAddressFamily: "
-			    "want ASN.1 sequence, have %s (NID %d)",
-			    p->fn, ASN1_tag2str(t->type), t->type);
+		switch (af->ipAddressChoice->type) {
+		case IPAddressChoice_inherit:
+			aors = NULL;
+			ipsz = p->res->ipsz + 1;
+			break;
+		case IPAddressChoice_addressesOrRanges:
+			aors = af->ipAddressChoice->u.addressesOrRanges;
+			ipsz = p->res->ipsz + sk_IPAddressOrRange_num(aors);
+			break;
+		default:
+			warnx("%s: RFC 3779: IPAddressChoice: unknown type %d",
+			    p->fn, af->ipAddressChoice->type);
 			goto out;
 		}
-		d = t->value.asn1_string->data;
-		dsz = t->value.asn1_string->length;
-		if (!sbgp_ipaddrfam(p, d, dsz))
+
+		if (ipsz >= MAX_IP_SIZE)
 			goto out;
+		p->res->ips = recallocarray(p->res->ips, p->res->ipsz, ipsz,
+		    sizeof(struct cert_ip));
+		if (p->res->ips == NULL)
+			err(1, NULL);
+
+		if (!ip_addr_afi_parse(p->fn, af->addressFamily, &afi)) {
+			warnx("%s: RFC 3779: invalid AFI", p->fn);
+			goto out;
+		}
+
+		if (aors == NULL) {
+			if (!sbgp_addr_inherit(p->fn, p->res->ips,
+			    &p->res->ipsz, afi))
+				goto out;
+			continue;
+		}
+
+		for (j = 0; j < sk_IPAddressOrRange_num(aors); j++) {
+			aor = sk_IPAddressOrRange_value(aors, j);
+			switch (aor->type) {
+			case IPAddressOrRange_addressPrefix:
+				if (!sbgp_addr(p->fn, p->res->ips,
+				    &p->res->ipsz, afi, aor->u.addressPrefix))
+					goto out;
+				break;
+			case IPAddressOrRange_addressRange:
+				if (!sbgp_addr_range(p->fn, p->res->ips,
+				    &p->res->ipsz, afi, aor->u.addressRange))
+					goto out;
+				break;
+			default:
+				warnx("%s: RFC 3779: IPAddressOrRange: "
+				    "unknown type %d", p->fn, aor->type);
+				goto out;
+			}
+		}
 	}
 
 	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	sk_ASN1_TYPE_pop_free(sseq, ASN1_TYPE_free);
-	free(sv);
+ out:
+	sk_IPAddressFamily_pop_free(addrblk, IPAddressFamily_free);
+	return rc;
+}
+
+/*
+ * Parse "Subject Information Access" extension, RFC 6487 4.8.8.
+ * Returns zero on failure, non-zero on success.
+ */
+static int
+sbgp_sia(struct parse *p, X509_EXTENSION *ext)
+{
+	AUTHORITY_INFO_ACCESS	*sia = NULL;
+	ACCESS_DESCRIPTION	*ad;
+	ASN1_OBJECT		*oid;
+	int			 i, rc = 0;
+
+	if (X509_EXTENSION_get_critical(ext)) {
+		warnx("%s: RFC 6487 section 4.8.8: SIA: "
+		    "extension not non-critical", p->fn);
+		goto out;
+	}
+
+	if ((sia = X509V3_EXT_d2i(ext)) == NULL) {
+		cryptowarnx("%s: RFC 6487 section 4.8.8: SIA: "
+		    "failed extension parse", p->fn);
+		goto out;
+	}
+
+	for (i = 0; i < sk_ACCESS_DESCRIPTION_num(sia); i++) {
+		ad = sk_ACCESS_DESCRIPTION_value(sia, i);
+
+		oid = ad->method;
+
+		if (OBJ_cmp(oid, carepo_oid) == 0) {
+			if (!x509_location(p->fn, "SIA: caRepository",
+			    "rsync://", ad->location, &p->res->repo))
+				goto out;
+		} else if (OBJ_cmp(oid, manifest_oid) == 0) {
+			if (!x509_location(p->fn, "SIA: rpkiManifest",
+			    "rsync://", ad->location, &p->res->mft))
+				goto out;
+		} else if (OBJ_cmp(oid, notify_oid) == 0) {
+			if (!x509_location(p->fn, "SIA: rpkiNotify",
+			    "https://", ad->location, &p->res->notify))
+				goto out;
+		}
+	}
+
+	if (p->res->mft == NULL || p->res->repo == NULL) {
+		warnx("%s: RFC 6487 section 4.8.8: SIA: missing caRepository "
+		    "or rpkiManifest", p->fn);
+		goto out;
+	}
+
+	if (strstr(p->res->mft, p->res->repo) != p->res->mft) {
+		warnx("%s: RFC 6487 section 4.8.8: SIA: "
+		    "conflicting URIs for caRepository and rpkiManifest",
+		    p->fn);
+		goto out;
+	}
+
+	if (rtype_from_file_extension(p->res->mft) != RTYPE_MFT) {
+		warnx("%s: RFC 6487 section 4.8.8: SIA: "
+		    "not an MFT file", p->fn);
+		goto out;
+	}
+
+	rc = 1;
+ out:
+	AUTHORITY_INFO_ACCESS_free(sia);
 	return rc;
 }
 
@@ -1065,7 +578,7 @@ certificate_policies(struct parse *p, X509_EXTENSION *ext)
 struct cert *
 cert_parse_pre(const char *fn, const unsigned char *der, size_t len)
 {
-	int		 rc = 0, extsz, c;
+	int		 extsz;
 	int		 sia_present = 0;
 	size_t		 i;
 	X509		*x = NULL;
@@ -1087,6 +600,12 @@ cert_parse_pre(const char *fn, const unsigned char *der, size_t len)
 		goto out;
 	}
 
+	/* Cache X509v3 extensions, see X509_check_ca(3). */
+	if (X509_check_purpose(x, -1, -1) <= 0) {
+		cryptowarnx("%s: could not cache X509v3 extensions", p.fn);
+		goto out;
+	}
+
 	/* Look for X509v3 extensions. */
 
 	if ((extsz = X509_get_ext_count(x)) < 0)
@@ -1097,21 +616,24 @@ cert_parse_pre(const char *fn, const unsigned char *der, size_t len)
 		assert(ext != NULL);
 		obj = X509_EXTENSION_get_object(ext);
 		assert(obj != NULL);
-		c = 1;
 
 		switch (OBJ_obj2nid(obj)) {
 		case NID_sbgp_ipAddrBlock:
-			c = sbgp_ipaddrblk(&p, ext);
+			if (!sbgp_ipaddrblk(&p, ext))
+				goto out;
 			break;
 		case NID_sbgp_autonomousSysNum:
-			c = sbgp_assysnum(&p, ext);
+			if (!sbgp_assysnum(&p, ext))
+				goto out;
 			break;
 		case NID_sinfo_access:
 			sia_present = 1;
-			c = sbgp_sia(&p, ext);
+			if (!sbgp_sia(&p, ext))
+				goto out;
 			break;
 		case NID_certificate_policies:
-			c = certificate_policies(&p, ext);
+			if (!certificate_policies(&p, ext))
+				goto out;
 			break;
 		case NID_crl_distribution_points:
 			/* ignored here, handled later */
@@ -1133,8 +655,6 @@ cert_parse_pre(const char *fn, const unsigned char *der, size_t len)
 			} */
 			break;
 		}
-		if (c == 0)
-			goto out;
 	}
 
 	if (!x509_get_aki(x, p.fn, &p.res->aki))
@@ -1170,12 +690,12 @@ cert_parse_pre(const char *fn, const unsigned char *der, size_t len)
 		}
 		if (p.res->ipsz > 0) {
 			warnx("%s: unexpected IP resources in BGPsec cert",
-			   p.fn);
+			    p.fn);
 			goto out;
 		}
 		if (sia_present) {
 			warnx("%s: unexpected SIA extension in BGPsec cert",
-			   p.fn);
+			    p.fn);
 			goto out;
 		}
 		break;
@@ -1190,19 +710,20 @@ cert_parse_pre(const char *fn, const unsigned char *der, size_t len)
 	}
 
 	p.res->x509 = x;
+	return p.res;
 
-	rc = 1;
 out:
-	if (rc == 0) {
-		cert_free(p.res);
-		X509_free(x);
-	}
-	return (rc == 0) ? NULL : p.res;
+	cert_free(p.res);
+	X509_free(x);
+	return NULL;
 }
 
 struct cert *
 cert_parse(const char *fn, struct cert *p)
 {
+	if (p == NULL)
+		return NULL;
+
 	if (p->aki == NULL) {
 		warnx("%s: RFC 6487 section 8.4.2: "
 		    "non-trust anchor missing AKI", fn);
@@ -1236,6 +757,9 @@ ta_parse(const char *fn, struct cert *p, const unsigned char *pkey,
 	ASN1_TIME	*notBefore, *notAfter;
 	EVP_PKEY	*pk, *opk;
 
+	if (p == NULL)
+		return NULL;
+
 	/* first check pubkey against the one from the TAL */
 	pk = d2i_PUBKEY(NULL, &pkey, pkeysz);
 	if (pk == NULL) {
@@ -1264,7 +788,7 @@ ta_parse(const char *fn, struct cert *p, const unsigned char *pkey,
 		warnx("%s: certificate not yet valid", fn);
 		goto badcert;
 	}
-	if (X509_cmp_current_time(notAfter) != 1)  {
+	if (X509_cmp_current_time(notAfter) != 1) {
 		warnx("%s: certificate has expired", fn);
 		goto badcert;
 	}
@@ -1390,6 +914,14 @@ cert_read(struct ibuf *b)
 	return p;
 }
 
+static inline int
+authcmp(struct auth *a, struct auth *b)
+{
+	return strcmp(a->cert->ski, b->cert->ski);
+}
+
+RB_GENERATE_STATIC(auth_tree, auth, entry, authcmp);
+
 struct auth *
 auth_find(struct auth_tree *auths, const char *aki)
 {
@@ -1403,7 +935,7 @@ auth_find(struct auth_tree *auths, const char *aki)
 	return RB_FIND(auth_tree, auths, &a);
 }
 
-void
+struct auth *
 auth_insert(struct auth_tree *auths, struct cert *cert, struct auth *parent)
 {
 	struct auth *na;
@@ -1417,16 +949,10 @@ auth_insert(struct auth_tree *auths, struct cert *cert, struct auth *parent)
 
 	if (RB_INSERT(auth_tree, auths, na) != NULL)
 		err(1, "auth tree corrupted");
-}
 
-static inline int
-authcmp(struct auth *a, struct auth *b)
-{
-	return strcmp(a->cert->ski, b->cert->ski);
+	return na;
 }
 
-RB_GENERATE(auth_tree, auth, entry, authcmp);
-
 static void
 insert_brk(struct brk_tree *tree, struct cert *cert, int asid)
 {
