@@ -1,4 +1,4 @@
-/*	$OpenBSD: cms.c,v 1.16 2022/03/28 13:04:01 claudio Exp $ */
+/*	$OpenBSD: cms.c,v 1.20 2022/05/31 18:41:43 tb Exp $ */
 /*
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -27,10 +27,10 @@
 
 #include "extern.h"
 
-extern ASN1_OBJECT     *cnt_type_oid;
-extern ASN1_OBJECT     *msg_dgst_oid;
-extern ASN1_OBJECT     *sign_time_oid;
-extern ASN1_OBJECT     *bin_sign_time_oid;
+extern ASN1_OBJECT	*cnt_type_oid;
+extern ASN1_OBJECT	*msg_dgst_oid;
+extern ASN1_OBJECT	*sign_time_oid;
+extern ASN1_OBJECT	*bin_sign_time_oid;
 
 /*
  * Parse and validate a self-signed CMS message, where the signing X509
@@ -224,6 +224,12 @@ cms_parse_validate(X509 **xp, const char *fn, const unsigned char *der,
 	}
 	*xp = X509_dup(sk_X509_value(certs, 0));
 
+	/* Cache X509v3 extensions, see X509_check_ca(3). */
+	if (X509_check_purpose(*xp, -1, -1) <= 0) {
+		cryptowarnx("%s: could not cache X509v3 extensions", fn);
+		goto out;
+	}
+
 	if (CMS_SignerInfo_get0_signer_id(si, &kid, NULL, NULL) != 1 ||
 	    kid == NULL) {
 		warnx("%s: RFC 6488: could not extract SKI from SID", fn);
@@ -266,61 +272,3 @@ out:
 
 	return res;
 }
-
-/*
- * Wrapper around ASN1_get_object() that preserves the current start
- * state and returns a more meaningful value.
- * Return zero on failure, non-zero on success.
- */
-int
-ASN1_frame(const char *fn, size_t sz,
-	const unsigned char **cnt, long *cntsz, int *tag)
-{
-	int	 ret, pcls;
-
-	ret = ASN1_get_object(cnt, cntsz, tag, &pcls, sz);
-	if ((ret & 0x80)) {
-		cryptowarnx("%s: ASN1_get_object", fn);
-		return 0;
-	}
-	return ASN1_object_size((ret & 0x01) ? 2 : 0, *cntsz, *tag);
-}
-
-/*
- * Check the version field in eContent.
- * Returns -1 on failure, zero on success.
- */
-int
-cms_econtent_version(const char *fn, const unsigned char **d, size_t dsz,
-	long *version)
-{
-	ASN1_INTEGER	*aint = NULL;
-	long		 plen;
-	int		 ptag, rc = -1;
-
-	if (!ASN1_frame(fn, dsz, d, &plen, &ptag))
-		goto out;
-	if (ptag != 0) {
-		warnx("%s: eContent version: expected explicit tag [0]", fn);
-		goto out;
-	}
-
-	aint = d2i_ASN1_INTEGER(NULL, d, plen);
-	if (aint == NULL) {
-		cryptowarnx("%s: eContent version: failed d2i_ASN1_INTEGER",
-		    fn);
-		goto out;
-	}
-
-	*version = ASN1_INTEGER_get(aint);
-	if (*version < 0) {
-		warnx("%s: eContent version: expected positive integer, got:"
-		    " %ld", fn, *version);
-		goto out;
-	}
-
-	rc = 0;
-out:
-	ASN1_INTEGER_free(aint);
-	return rc;
-}
