@@ -1,4 +1,4 @@
-/*	$OpenBSD: extern.h,v 1.125 2022/04/04 16:02:54 claudio Exp $ */
+/*	$OpenBSD: extern.h,v 1.143 2022/06/27 10:18:27 job Exp $ */
 /*
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -22,6 +22,24 @@
 #include <sys/time.h>
 
 #include <openssl/x509.h>
+#include <openssl/x509v3.h>
+
+/*
+ * Distrusted hosts (loaded from skipfile).
+ */
+struct skiplistentry {
+	LIST_ENTRY(skiplistentry)	 entry;
+	char				*value; /* FQDN */
+};
+LIST_HEAD(skiplist, skiplistentry);
+
+/*
+ * Enumeration for ASN.1 explicit tags in RSC eContent
+ */
+enum rsc_resourceblock_tag {
+	RSRCBLK_TYPE_ASID,
+	RSRCBLK_TYPE_IPADDRBLK,
+};
 
 enum cert_as_type {
 	CERT_AS_ID, /* single identifier */
@@ -163,6 +181,7 @@ enum rtype {
 	RTYPE_ASPA,
 	RTYPE_REPO,
 	RTYPE_FILE,
+	RTYPE_RSC,
 };
 
 enum location {
@@ -193,7 +212,9 @@ struct mft {
 	char		*aia; /* AIA */
 	char		*aki; /* AKI */
 	char		*ski; /* SKI */
-	time_t		 valid_from;
+	char		*crl; /* CRL file name */
+	unsigned char	 crlhash[SHA256_DIGEST_LENGTH];
+	time_t		 valid_since;
 	time_t		 valid_until;
 	size_t		 filesz; /* number of filenames */
 	unsigned int	 repoid;
@@ -229,6 +250,29 @@ struct roa {
 	time_t		 expires; /* do not use after */
 };
 
+struct rscfile {
+	char		*filename; /* an optional filename on the checklist */
+	unsigned char	 hash[SHA256_DIGEST_LENGTH]; /* the digest */
+};
+
+/*
+ * A Signed Checklist (RSC)
+ */
+struct rsc {
+	int		 talid; /* RSC covered by what TAL */
+	int		 valid; /* eContent resources covered by EE's 3779? */
+	struct cert_ip	*ips; /* IP prefixes */
+	size_t		 ipsz; /* number of IP prefixes */
+	struct cert_as	*as; /* AS resources */
+	size_t		 asz; /* number of AS resources */
+	struct rscfile	*files; /* FileAndHashes in the RSC */
+	size_t		 filesz; /* number of FileAndHashes */
+	char		*aia; /* AIA */
+	char		*aki; /* AKI */
+	char		*ski; /* SKI */
+	time_t		 expires; /* Not After of the RSC EE */
+};
+
 /*
  * A single Ghostbuster record
  */
@@ -288,7 +332,6 @@ struct crl {
  * Tree of CRLs sorted by uri
  */
 RB_HEAD(crl_tree, crl);
-RB_PROTOTYPE(crl_tree, crl, entry, crlcmp);
 
 /*
  * An authentication tuple.
@@ -304,10 +347,9 @@ struct auth {
  * Tree of auth sorted by ski
  */
 RB_HEAD(auth_tree, auth);
-RB_PROTOTYPE(auth_tree, auth, entry, authcmp);
 
 struct auth	*auth_find(struct auth_tree *, const char *);
-void		 auth_insert(struct auth_tree *, struct cert *, struct auth *);
+struct auth	*auth_insert(struct auth_tree *, struct cert *, struct auth *);
 
 enum http_result {
 	HTTP_FAILED,	/* anything else */
@@ -356,7 +398,7 @@ struct entity {
 	char		*path;		/* path relative to repository */
 	char		*file;		/* filename or valid repo path */
 	unsigned char	*data;		/* optional data blob */
-	size_t		 datasz; 	/* length of optional data blob */
+	size_t		 datasz;	/* length of optional data blob */
 	unsigned int	 repoid;	/* repository identifier */
 	int		 talid;		/* tal identifier */
 	enum rtype	 type;		/* type of entity (not RTYPE_EOF) */
@@ -397,6 +439,7 @@ struct stats {
 	size_t	 extra_files; /* number of superfluous files */
 	size_t	 del_dirs; /* number of directories removed in cleanup */
 	size_t	 brks; /* number of BGPsec Router Key (BRK) certificates */
+	size_t	 skiplistentries; /* number of skiplist entries */
 	struct timeval	elapsed_time;
 	struct timeval	user_time;
 	struct timeval	system_time;
@@ -411,7 +454,7 @@ extern int filemode;
 extern const char *tals[];
 extern const char *taldescs[];
 extern unsigned int talrepocnt[];
-extern size_t talsz;
+extern int talsz;
 
 /* Routines for RPKI entities. */
 
@@ -449,8 +492,14 @@ void		 gbr_free(struct gbr *);
 struct gbr	*gbr_parse(X509 **, const char *, const unsigned char *,
 		    size_t);
 
+void		 rsc_free(struct rsc *);
+struct rsc	*rsc_parse(X509 **, const char *, const unsigned char *,
+		    size_t);
+
 /* crl.c */
 struct crl	*crl_parse(const char *, const unsigned char *, size_t);
+struct crl	*crl_get(struct crl_tree *, const struct auth *);
+int		 crl_insert(struct crl_tree *, struct crl *);
 void		 crl_free(struct crl *);
 
 /* Validation of our objects. */
@@ -462,18 +511,19 @@ int		 valid_ta(const char *, struct auth_tree *,
 int		 valid_cert(const char *, struct auth *, const struct cert *);
 int		 valid_roa(const char *, struct auth *, struct roa *);
 int		 valid_filehash(int, const char *, size_t);
+int		 valid_hash(unsigned char *, size_t, const char *, size_t);
+int		 valid_filename(const char *, size_t);
 int		 valid_uri(const char *, size_t, const char *);
 int		 valid_origin(const char *, const char *);
+int		 valid_x509(char *, X509_STORE_CTX *, X509 *, struct auth *,
+		    struct crl *, int);
+int		 valid_rsc(const char *, struct auth *, struct rsc *);
+int		 valid_econtent_version(const char *, const ASN1_INTEGER *);
 
 /* Working with CMS. */
 unsigned char	*cms_parse_validate(X509 **, const char *,
 		    const unsigned char *, size_t,
 		    const ASN1_OBJECT *, size_t *);
-int		 cms_econtent_version(const char *, const unsigned char **,
-		    size_t, long *);
-/* Helper for ASN1 parsing */
-int		 ASN1_frame(const char *, size_t,
-			const unsigned char **, long *, int *);
 
 /* Work with RFC 3779 IP addresses, prefixes, ranges. */
 
@@ -491,6 +541,11 @@ int		 ip_addr_check_covered(enum afi, const unsigned char *,
 int		 ip_cert_compose_ranges(struct cert_ip *);
 void		 ip_roa_compose_ranges(struct roa_ip *);
 
+int		 sbgp_addr(const char *, struct cert_ip *, size_t *,
+		    enum afi, const ASN1_BIT_STRING *);
+int		 sbgp_addr_range(const char *, struct cert_ip *, size_t *,
+		    enum afi, const IPAddressRange *);
+
 /* Work with RFC 3779 AS numbers, ranges. */
 
 int		 as_id_parse(const ASN1_INTEGER *, uint32_t *);
@@ -499,11 +554,17 @@ int		 as_check_overlap(const struct cert_as *, const char *,
 int		 as_check_covered(uint32_t, uint32_t,
 			const struct cert_as *, size_t);
 
+int		 sbgp_as_id(const char *, struct cert_as *, size_t *,
+		    const ASN1_INTEGER *);
+int		 sbgp_as_range(const char *, struct cert_as *, size_t *,
+		    const ASRange *);
+
 /* Parser-specific */
 void		 entity_free(struct entity *);
 void		 entity_read_req(struct ibuf *, struct entity *);
 void		 entityq_flush(struct entityq *, struct repo *);
 void		 proc_parser(int) __attribute__((noreturn));
+void		 proc_filemode(int) __attribute__((noreturn));
 
 /* Rsync-specific. */
 
@@ -512,8 +573,8 @@ void		 proc_rsync(char *, char *, int) __attribute__((noreturn));
 
 /* HTTP and RRDP processes. */
 
-void		 proc_http(char *, int);
-void		 proc_rrdp(int);
+void		 proc_http(char *, int) __attribute__((noreturn));
+void		 proc_rrdp(int) __attribute__((noreturn));
 
 /* Repository handling */
 int		 filepath_add(struct filepath_tree *, char *);
@@ -589,15 +650,20 @@ char		*x509_get_pubkey(X509 *, const char *);
 enum cert_purpose	 x509_get_purpose(X509 *, const char *);
 int		 x509_get_time(const ASN1_TIME *, time_t *);
 char		*x509_convert_seqnum(const char *, const ASN1_INTEGER *);
+int		 x509_location(const char *, const char *, const char *,
+		    GENERAL_NAME *, char **);
+int		 x509_inherits(X509 *);
 
 /* printers */
 char		*time2str(time_t);
+void		 x509_print(const X509 *);
 void		 tal_print(const struct tal *);
 void		 cert_print(const struct cert *);
 void		 crl_print(const struct crl *);
-void		 mft_print(const struct mft *);
-void		 roa_print(const struct roa *);
-void		 gbr_print(const struct gbr *);
+void		 mft_print(const X509 *, const struct mft *);
+void		 roa_print(const X509 *, const struct roa *);
+void		 gbr_print(const X509 *, const struct gbr *);
+void		 rsc_print(const X509 *, const struct rsc *);
 
 /* Output! */
 
@@ -650,6 +716,8 @@ int	mkpathat(int, const char *);
 #define RPKI_RSYNC_CMD		"openrsync"
 #endif
 
+#define DEFAULT_SKIPLIST_FILE	RPKI_PATH_TAL_DIR "/skiplist"
+
 /* Maximum number of TAL files we'll load. */
 #define	TALSZ_MAX	8
 
@@ -663,14 +731,18 @@ int	mkpathat(int, const char *);
 /* Maximum acceptable file size */
 #define MAX_FILE_SIZE		4000000
 
+/* Maximum number of FileNameAndHash entries per RSC checklist. */
+#define MAX_CHECKLIST_ENTRIES	100000
+
 /* Maximum number of FileAndHash entries per manifest. */
 #define MAX_MANIFEST_ENTRIES	100000
 
 /* Maximum depth of the RPKI tree. */
 #define MAX_CERT_DEPTH		12
 
-/* Maximum number of concurrent rsync processes. */
-#define MAX_RSYNC_PROCESSES	16
+/* Maximum number of concurrent http and rsync requests. */
+#define MAX_HTTP_REQUESTS	64
+#define MAX_RSYNC_REQUESTS	16
 
 /* Maximum allowd repositories per tal */
 #define MAX_REPO_PER_TAL	1000
