@@ -1,4 +1,4 @@
-/*	$OpenBSD: main.c,v 1.192 2022/04/04 16:02:54 claudio Exp $ */
+/*	$OpenBSD: main.c,v 1.208 2022/06/27 10:18:27 job Exp $ */
 /*
  * Copyright (c) 2021 Claudio Jeker <claudio@openbsd.org>
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
@@ -25,6 +25,7 @@
 #include <sys/wait.h>
 
 #include <assert.h>
+#include <ctype.h>
 #include <err.h>
 #include <errno.h>
 #include <dirent.h>
@@ -47,7 +48,7 @@
 const char	*tals[TALSZ_MAX];
 const char	*taldescs[TALSZ_MAX];
 unsigned int	 talrepocnt[TALSZ_MAX];
-size_t		 talsz;
+int		 talsz;
 
 size_t	entity_queue;
 int	timeout = 60*60;
@@ -66,6 +67,8 @@ int	filemode;
 int	rrdpon = 1;
 int	repo_timeout;
 
+struct skiplist skiplist = LIST_HEAD_INITIALIZER(skiplist);
+
 struct stats	 stats;
 
 /*
@@ -330,51 +333,25 @@ rrdp_http_done(unsigned int id, enum http_result res, const char *last_mod)
  * These are always relative to the directory in which "mft" sits.
  */
 static void
-queue_add_from_mft(const char *path, const struct mftfile *file,
-    struct repo *rp)
-{
-	char		*nfile, *npath = NULL;
-
-	if (path != NULL)
-		if ((npath = strdup(path)) == NULL)
-			err(1, NULL);
-	if ((nfile = strdup(file->file)) == NULL)
-		err(1, NULL);
-
-	entityq_add(npath, nfile, file->type, file->location, rp, NULL, 0, -1);
-}
-
-/*
- * Loops over queue_add_from_mft() for all files.
- * The order here is important: we want to parse the revocation
- * list *before* we parse anything else.
- */
-static void
-queue_add_from_mft_set(const struct mft *mft, const char *name, struct repo *rp)
+queue_add_from_mft(const struct mft *mft, struct repo *rp)
 {
 	size_t			 i;
 	const struct mftfile	*f;
+	char			*nfile, *npath = NULL;
 
 	for (i = 0; i < mft->filesz; i++) {
 		f = &mft->files[i];
-		if (f->type != RTYPE_CRL)
-			continue;
-		queue_add_from_mft(mft->path, f, rp);
-	}
 
-	for (i = 0; i < mft->filesz; i++) {
-		f = &mft->files[i];
-		switch (f->type) {
-		case RTYPE_CER:
-		case RTYPE_ROA:
-		case RTYPE_GBR:
-			queue_add_from_mft(mft->path, f, rp);
-			break;
-		case RTYPE_CRL:
+		if (f->type == RTYPE_INVALID)
 			continue;
-		default:
-			warnx("%s: unsupported file: %s", name, f->file);
-		}
+
+		if (mft->path != NULL)
+			if ((npath = strdup(mft->path)) == NULL)
+				err(1, NULL);
+		if ((nfile = strdup(f->file)) == NULL)
+			err(1, NULL);
+		entityq_add(npath, nfile, f->type, f->location, rp, NULL, 0,
+		    -1);
 	}
 }
 
@@ -441,10 +418,22 @@ queue_add_from_tal(struct tal *tal)
 static void
 queue_add_from_cert(const struct cert *cert)
 {
-	struct repo	*repo;
-	char		*nfile, *npath;
-	const char	*uri, *repouri, *file;
-	size_t		 repourisz;
+	struct repo		*repo;
+	struct skiplistentry	*sle;
+	char			*nfile, *npath, *host;
+	const char		*uri, *repouri, *file;
+	size_t			 repourisz;
+
+	LIST_FOREACH(sle, &skiplist, entry) {
+		if (strncmp(cert->repo, "rsync://", 8) != 0)
+			errx(1, "unexpected protocol");
+		host = cert->repo + 8;
+
+		if (strncasecmp(host, sle->value, strcspn(host, "/")) == 0) {
+			warnx("skipping %s (listed in skiplist)", cert->repo);
+			return;
+		}
+	}
 
 	repo = repo_lookup(cert->talid, cert->repo,
 	    rrdpon ? cert->notify : NULL);
@@ -529,19 +518,18 @@ entity_process(struct ibuf *b, struct stats *st, struct vrp_tree *tree,
 			break;
 		}
 		cert = cert_read(b);
-		if (cert->purpose == CERT_PURPOSE_CA) {
-			/*
-			 * Process the revocation list from the
-			 * certificate *first*, since it might mark that
-			 * we're revoked and then we don't want to
-			 * process the MFT.
-			 */
+		switch (cert->purpose) {
+		case CERT_PURPOSE_CA:
 			queue_add_from_cert(cert);
-		} else if (cert->purpose == CERT_PURPOSE_BGPSEC_ROUTER) {
+			break;
+		case CERT_PURPOSE_BGPSEC_ROUTER:
 			cert_insert_brks(brktree, cert);
 			st->brks++;
-		} else
+			break;
+		default:
 			st->certs_fail++;
+			break;
+		}
 		cert_free(cert);
 		break;
 	case RTYPE_MFT:
@@ -553,8 +541,7 @@ entity_process(struct ibuf *b, struct stats *st, struct vrp_tree *tree,
 		}
 		mft = mft_read(b);
 		if (!mft->stale)
-			queue_add_from_mft_set(mft, file,
-			    repo_byid(mft->repoid));
+			queue_add_from_mft(mft, repo_byid(mft->repoid));
 		else
 			st->mfts_stale++;
 		mft_free(mft);
@@ -649,13 +636,13 @@ rrdp_process(struct ibuf *b)
  * Assign filenames ending in ".tal" in RPKI_PATH_TAL_DIR into "tals",
  * returning the number of files found and filled-in.
  * This may be zero.
- * Don't exceded "max" filenames.
+ * Don't exceed "max" filenames.
  */
-static size_t
+static int
 tal_load_default(void)
 {
 	static const char *confdir = RPKI_PATH_TAL_DIR;
-	size_t s = 0;
+	int s = 0;
 	char *path;
 	DIR *dirp;
 	struct dirent *dp;
@@ -677,6 +664,57 @@ tal_load_default(void)
 	return s;
 }
 
+/*
+ * Load the list of FQDNs from the skiplist which are to be distrusted.
+ * Return 0 on success.
+ */
+static void
+load_skiplist(const char *slf)
+{
+	struct skiplistentry	*sle;
+	FILE			*fp;
+	char			*line = NULL;
+	size_t			 linesize = 0, s;
+	ssize_t			 linelen;
+
+	if ((fp = fopen(slf, "r")) == NULL) {
+		if (strcmp(slf, DEFAULT_SKIPLIST_FILE) != 0)
+			errx(1, "failed to open skiplist %s", slf);
+		return;
+	}
+
+	while ((linelen = getline(&line, &linesize, fp)) != -1) {
+		/* just eat comment lines or empty lines*/
+		if (line[0] == '#' || line[0] == '\n')
+			continue;
+
+		if (line[0] == ' ' || line[0] == '\t')
+			errx(1, "invalid entry in skiplist: %s", line);
+
+		/*
+		 * Ignore anything after comment sign, whitespaces,
+		 * also chop off LF or CR.
+		 */
+		line[strcspn(line, " #\r\n\t")] = 0;
+
+		for (s = 0; s < strlen(line); s++)
+			if (!isalnum((unsigned char)line[s]) &&
+			    !ispunct((unsigned char)line[s]))
+				errx(1, "invalid entry in skiplist: %s", line);
+
+		if ((sle = malloc(sizeof(struct skiplistentry))) == NULL)
+			err(1, NULL);
+		if ((sle->value = strdup(line)) == NULL)
+			err(1, NULL);
+
+		LIST_INSERT_HEAD(&skiplist, sle, entry);
+		stats.skiplistentries++;
+	}
+
+	fclose(fp);
+	free(line);
+}
+
 static void
 check_fs_size(int fd, const char *cachedir)
 {
@@ -703,6 +741,34 @@ check_fs_size(int fd, const char *cachedir)
 	}
 }
 
+static pid_t
+process_start(const char *title, int *fd)
+{
+	int		 fl = SOCK_STREAM | SOCK_CLOEXEC | SOCK_NONBLOCK;
+	pid_t		 pid;
+	int		 pair[2];
+
+	if (socketpair(AF_UNIX, fl, 0, pair) == -1)
+		err(1, "socketpair");
+	if ((pid = fork()) == -1)
+		err(1, "fork");
+
+	if (pid == 0) {
+		setproctitle("%s", title);
+		/* change working directory to the cache directory */
+		if (fchdir(cachefd) == -1)
+			err(1, "fchdir");
+		if (!filemode && timeout > 0)
+			alarm(timeout);
+		close(pair[1]);
+		*fd = pair[0];
+	} else {
+		close(pair[0]);
+		*fd = pair[1];
+	}
+	return pid;
+}
+
 void
 suicide(int sig __attribute__((unused)))
 {
@@ -714,11 +780,8 @@ suicide(int sig __attribute__((unused)))
 int
 main(int argc, char *argv[])
 {
-	int		 rc, c, st, proc, rsync, http, rrdp, hangup = 0;
-	int		 fl = SOCK_STREAM | SOCK_CLOEXEC | SOCK_NONBLOCK;
-	size_t		 i;
+	int		 rc, c, i, st, proc, rsync, http, rrdp, hangup = 0;
 	pid_t		 pid, procpid, rsyncpid, httppid, rrdppid;
-	int		 fd[2];
 	struct pollfd	 pfd[NPFD];
 	struct msgbuf	*queues[NPFD];
 	struct ibuf	*b, *httpbuf = NULL, *procbuf = NULL;
@@ -727,10 +790,11 @@ main(int argc, char *argv[])
 	char		*bind_addr = NULL;
 	const char	*cachedir = NULL, *outputdir = NULL;
 	const char	*errs, *name;
+	const char	*skiplistfile = NULL;
 	struct vrp_tree	 vrps = RB_INITIALIZER(&vrps);
-	struct brk_tree  brks = RB_INITIALIZER(&brks);
-	struct rusage	ru;
-	struct timeval	start_time, now_time;
+	struct brk_tree	 brks = RB_INITIALIZER(&brks);
+	struct rusage	 ru;
+	struct timeval	 start_time, now_time;
 
 	gettimeofday(&start_time, NULL);
 
@@ -749,12 +813,13 @@ main(int argc, char *argv[])
 	cachedir = RPKI_PATH_BASE_DIR;
 	outputdir = RPKI_PATH_OUT_DIR;
 	repo_timeout = timeout / 4;
+	skiplistfile = DEFAULT_SKIPLIST_FILE;
 
 	if (pledge("stdio rpath wpath cpath inet fattr dns sendfd recvfd "
 	    "proc exec unveil", NULL) == -1)
 		err(1, "pledge");
 
-	while ((c = getopt(argc, argv, "b:Bcd:e:fjnorRs:t:T:vV")) != -1)
+	while ((c = getopt(argc, argv, "b:Bcd:e:fjnorRs:S:t:T:vV")) != -1)
 		switch (c) {
 		case 'b':
 			bind_addr = optarg;
@@ -796,15 +861,15 @@ main(int argc, char *argv[])
 				errx(1, "-s: %s", errs);
 			if (timeout == 0)
 				repo_timeout = 24*60*60;
-			else if (timeout < 1)
-				errx(1, "-s: %i too small", timeout);
 			else
 				repo_timeout = timeout / 4;
 			break;
+		case 'S':
+			skiplistfile = optarg;
+			break;
 		case 't':
 			if (talsz >= TALSZ_MAX)
-				err(1,
-				    "too many tal files specified");
+				err(1, "too many tal files specified");
 			tals[talsz++] = optarg;
 			break;
 		case 'T':
@@ -853,15 +918,12 @@ main(int argc, char *argv[])
 		if ((outdirfd = open(outputdir, O_RDONLY | O_DIRECTORY)) == -1)
 			err(1, "output directory %s", outputdir);
 		if (outformats == 0)
-			outformats = FORMAT_OPENBGPD;
+			outformats = FORMAT_OPENBGPD | FORMAT_JSON |
+			    FORMAT_CSV | FORMAT_BIRD;
 	}
 
 	check_fs_size(cachefd, cachedir);
 
-	if (outformats == 0)
-		outformats = FORMAT_OPENBGPD | FORMAT_JSON | FORMAT_CSV
-		    | FORMAT_BIRD;
-
 	if (talsz == 0)
 		talsz = tal_load_default();
 	if (talsz == 0)
@@ -873,34 +935,14 @@ main(int argc, char *argv[])
 	 * manifests, certificates, etc.) and returning contents.
 	 */
 
-	if (socketpair(AF_UNIX, fl, 0, fd) == -1)
-		err(1, "socketpair");
-	if ((procpid = fork()) == -1)
-		err(1, "fork");
-
+	procpid = process_start("parser", &proc);
 	if (procpid == 0) {
-		close(fd[1]);
-
-		setproctitle("parser");
-		/* change working directory to the cache directory */
-		if (fchdir(cachefd) == -1)
-			err(1, "fchdir");
-
-		if (timeout)
-			alarm(timeout);
-
-		/* Only allow access to the cache directory. */
-		if (unveil(".", "r") == -1)
-			err(1, "%s: unveil", cachedir);
-		if (pledge("stdio rpath", NULL) == -1)
-			err(1, "pledge");
-		proc_parser(fd[0]);
-		errx(1, "parser process returned");
+		if (!filemode)
+			proc_parser(proc);
+		else
+			proc_filemode(proc);
 	}
 
-	close(fd[0]);
-	proc = fd[1];
-
 	/*
 	 * Create a process that will do the rsync'ing.
 	 * This process is responsible for making sure that all the
@@ -909,32 +951,11 @@ main(int argc, char *argv[])
 	 */
 
 	if (!noop) {
-		if (socketpair(AF_UNIX, fl, 0, fd) == -1)
-			err(1, "socketpair");
-		if ((rsyncpid = fork()) == -1)
-			err(1, "fork");
-
+		rsyncpid = process_start("rsync", &rsync);
 		if (rsyncpid == 0) {
 			close(proc);
-			close(fd[1]);
-
-			setproctitle("rsync");
-			/* change working directory to the cache directory */
-			if (fchdir(cachefd) == -1)
-				err(1, "fchdir");
-
-			if (timeout)
-				alarm(timeout);
-
-			if (pledge("stdio rpath proc exec unveil", NULL) == -1)
-				err(1, "pledge");
-
-			proc_rsync(rsync_prog, bind_addr, fd[0]);
-			errx(1, "rsync process returned");
+			proc_rsync(rsync_prog, bind_addr, rsync);
 		}
-
-		close(fd[0]);
-		rsync = fd[1];
 	} else {
 		rsync = -1;
 		rsyncpid = -1;
@@ -946,34 +967,14 @@ main(int argc, char *argv[])
 	 * where the data should be written to.
 	 */
 
-	if (!noop) {
-		if (socketpair(AF_UNIX, fl, 0, fd) == -1)
-			err(1, "socketpair");
-		if ((httppid = fork()) == -1)
-			err(1, "fork");
+	if (!noop && rrdpon) {
+		httppid = process_start("http", &http);
 
 		if (httppid == 0) {
 			close(proc);
 			close(rsync);
-			close(fd[1]);
-
-			setproctitle("http");
-			/* change working directory to the cache directory */
-			if (fchdir(cachefd) == -1)
-				err(1, "fchdir");
-
-			if (timeout)
-				alarm(timeout);
-
-			if (pledge("stdio rpath inet dns recvfd", NULL) == -1)
-				err(1, "pledge");
-
-			proc_http(bind_addr, fd[0]);
-			errx(1, "http process returned");
+			proc_http(bind_addr, http);
 		}
-
-		close(fd[0]);
-		http = fd[1];
 	} else {
 		http = -1;
 		httppid = -1;
@@ -986,40 +987,19 @@ main(int argc, char *argv[])
 	 */
 
 	if (!noop && rrdpon) {
-		if (socketpair(AF_UNIX, fl, 0, fd) == -1)
-			err(1, "socketpair");
-		if ((rrdppid = fork()) == -1)
-			err(1, "fork");
-
+		rrdppid = process_start("rrdp", &rrdp);
 		if (rrdppid == 0) {
 			close(proc);
 			close(rsync);
 			close(http);
-			close(fd[1]);
-
-			setproctitle("rrdp");
-			/* change working directory to the cache directory */
-			if (fchdir(cachefd) == -1)
-				err(1, "fchdir");
-
-			if (timeout)
-				alarm(timeout);
-
-			if (pledge("stdio recvfd", NULL) == -1)
-				err(1, "pledge");
-
-			proc_rrdp(fd[0]);
-			/* NOTREACHED */
+			proc_rrdp(rrdp);
 		}
-
-		close(fd[0]);
-		rrdp = fd[1];
 	} else {
 		rrdp = -1;
 		rrdppid = -1;
 	}
 
-	if (timeout) {
+	if (!filemode && timeout > 0) {
 		/*
 		 * Commit suicide eventually
 		 * cron will normally start a new one
@@ -1056,9 +1036,11 @@ main(int argc, char *argv[])
 	pfd[3].fd = rrdp;
 	queues[3] = &rrdpq;
 
+	load_skiplist(skiplistfile);
+
 	/*
-	 * Prime the process with our TAL file.
-	 * This will contain (hopefully) links to our manifest and we
+	 * Prime the process with our TAL files.
+	 * These will (hopefully) contain links to manifests and we
 	 * can get the ball rolling.
 	 */
 
@@ -1093,7 +1075,7 @@ main(int argc, char *argv[])
 
 		for (i = 0; i < NPFD; i++) {
 			if (pfd[i].revents & (POLLERR|POLLNVAL)) {
-				warnx("poll[%zu]: bad fd", i);
+				warnx("poll[%d]: bad fd", i);
 				hangup = 1;
 			}
 			if (pfd[i].revents & POLLHUP)
@@ -1101,12 +1083,12 @@ main(int argc, char *argv[])
 			if (pfd[i].revents & POLLOUT) {
 				switch (msgbuf_write(queues[i])) {
 				case 0:
-					warnx("write[%zu]: "
+					warnx("write[%d]: "
 					    "connection closed", i);
 					hangup = 1;
 					break;
 				case -1:
-					warn("write[%zu]", i);
+					warn("write[%d]", i);
 					hangup = 1;
 					break;
 				}
@@ -1256,26 +1238,27 @@ main(int argc, char *argv[])
 	if (outputfiles(&vrps, &brks, &stats))
 		rc = 1;
 
-	logx("Processing time %lld seconds "
-	    "(%lld seconds user, %lld seconds system)",
+	printf("Processing time %lld seconds "
+	    "(%lld seconds user, %lld seconds system)\n",
 	    (long long)stats.elapsed_time.tv_sec,
 	    (long long)stats.user_time.tv_sec,
 	    (long long)stats.system_time.tv_sec);
-	logx("Route Origin Authorizations: %zu (%zu failed parse, %zu invalid)",
+	printf("Skiplist entries: %zu\n", stats.skiplistentries);
+	printf("Route Origin Authorizations: %zu (%zu failed parse, %zu invalid)\n",
 	    stats.roas, stats.roas_fail, stats.roas_invalid);
-	logx("BGPsec Router Certificates: %zu", stats.brks);
-	logx("Certificates: %zu (%zu invalid)",
+	printf("BGPsec Router Certificates: %zu\n", stats.brks);
+	printf("Certificates: %zu (%zu invalid)\n",
 	    stats.certs, stats.certs_fail);
-	logx("Trust Anchor Locators: %zu (%zu invalid)",
+	printf("Trust Anchor Locators: %zu (%zu invalid)\n",
 	    stats.tals, talsz - stats.tals);
-	logx("Manifests: %zu (%zu failed parse, %zu stale)",
+	printf("Manifests: %zu (%zu failed parse, %zu stale)\n",
 	    stats.mfts, stats.mfts_fail, stats.mfts_stale);
-	logx("Certificate revocation lists: %zu", stats.crls);
-	logx("Ghostbuster records: %zu", stats.gbrs);
-	logx("Repositories: %zu", stats.repos);
-	logx("Cleanup: removed %zu files, %zu directories, %zu superfluous",
+	printf("Certificate revocation lists: %zu\n", stats.crls);
+	printf("Ghostbuster records: %zu\n", stats.gbrs);
+	printf("Repositories: %zu\n", stats.repos);
+	printf("Cleanup: removed %zu files, %zu directories, %zu superfluous\n",
 	    stats.del_files, stats.del_dirs, stats.extra_files);
-	logx("VRP Entries: %zu (%zu unique)", stats.vrps, stats.uniqs);
+	printf("VRP Entries: %zu (%zu unique)\n", stats.vrps, stats.uniqs);
 
 	/* Memory cleanup. */
 	repo_free();
@@ -1286,7 +1269,7 @@ usage:
 	fprintf(stderr,
 	    "usage: rpki-client [-BcjnoRrVv] [-b sourceaddr] [-d cachedir]"
 	    " [-e rsync_prog]\n"
-	    "                   [-s timeout] [-T table] [-t tal]"
+	    "                   [-S skiplist] [-s timeout] [-T table] [-t tal]"
 	    " [outputdir]\n"
 	    "       rpki-client [-Vv] [-d cachedir] [-t tal] -f file ...\n");
 	return 1;
