@@ -1,5 +1,6 @@
-/*	$OpenBSD: mft.c,v 1.55 2022/04/01 17:22:07 claudio Exp $ */
+/*	$OpenBSD: mft.c,v 1.72 2022/06/10 10:41:09 tb Exp $ */
 /*
+ * Copyright (c) 2022 Theo Buehler <tb@openbsd.org>
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -27,7 +28,10 @@
 
 #include <openssl/bn.h>
 #include <openssl/asn1.h>
+#include <openssl/asn1t.h>
+#include <openssl/safestack.h>
 #include <openssl/sha.h>
+#include <openssl/stack.h>
 #include <openssl/x509.h>
 
 #include "extern.h"
@@ -38,9 +42,52 @@
 struct	parse {
 	const char	*fn; /* manifest file name */
 	struct mft	*res; /* result object */
+	int		 found_crl;
 };
 
-extern ASN1_OBJECT    *mft_oid;
+extern ASN1_OBJECT	*mft_oid;
+
+/*
+ * Types and templates for the Manifest eContent, RFC 6486, section 4.2.
+ */
+
+typedef struct {
+	ASN1_IA5STRING	*file;
+	ASN1_BIT_STRING	*hash;
+} FileAndHash;
+
+DECLARE_STACK_OF(FileAndHash);
+
+#ifndef DEFINE_STACK_OF
+#define sk_FileAndHash_num(sk)		SKM_sk_num(FileAndHash, (sk))
+#define sk_FileAndHash_value(sk, i)	SKM_sk_value(FileAndHash, (sk), (i))
+#endif
+
+typedef struct {
+	ASN1_INTEGER		*version;
+	ASN1_INTEGER		*manifestNumber;
+	ASN1_GENERALIZEDTIME	*thisUpdate;
+	ASN1_GENERALIZEDTIME	*nextUpdate;
+	ASN1_OBJECT		*fileHashAlg;
+	STACK_OF(FileAndHash)	*fileList;
+} Manifest;
+
+ASN1_SEQUENCE(FileAndHash) = {
+	ASN1_SIMPLE(FileAndHash, file, ASN1_IA5STRING),
+	ASN1_SIMPLE(FileAndHash, hash, ASN1_BIT_STRING),
+} ASN1_SEQUENCE_END(FileAndHash);
+
+ASN1_SEQUENCE(Manifest) = {
+	ASN1_IMP_OPT(Manifest, version, ASN1_INTEGER, 0),
+	ASN1_SIMPLE(Manifest, manifestNumber, ASN1_INTEGER),
+	ASN1_SIMPLE(Manifest, thisUpdate, ASN1_GENERALIZEDTIME),
+	ASN1_SIMPLE(Manifest, nextUpdate, ASN1_GENERALIZEDTIME),
+	ASN1_SIMPLE(Manifest, fileHashAlg, ASN1_OBJECT),
+	ASN1_SEQUENCE_OF(Manifest, fileList, FileAndHash),
+} ASN1_SEQUENCE_END(Manifest);
+
+DECLARE_ASN1_FUNCTIONS(Manifest);
+IMPLEMENT_ASN1_FUNCTIONS(Manifest);
 
 /*
  * Convert an ASN1_GENERALIZEDTIME to a struct tm.
@@ -85,9 +132,9 @@ mft_parse_time(const ASN1_GENERALIZEDTIME *from,
 		return 0;
 	}
 
-	if ((p->res->valid_from = mktime(&tm_from)) == -1 ||
-	    (p->res->valid_until = mktime(&tm_until)) == -1)
-		errx(1, "%s: mktime failed", p->fn);
+	if ((p->res->valid_since = timegm(&tm_from)) == -1 ||
+	    (p->res->valid_until = timegm(&tm_until)) == -1)
+		errx(1, "%s: timegm failed", p->fn);
 
 	return 1;
 }
@@ -119,6 +166,8 @@ rtype_from_file_extension(const char *fn)
 		return RTYPE_GBR;
 	if (strcasecmp(fn + sz - 4, ".asa") == 0)
 		return RTYPE_ASPA;
+	if (strcasecmp(fn + sz - 4, ".sig") == 0)
+		return RTYPE_RSC;
 
 	return RTYPE_INVALID;
 }
@@ -126,16 +175,15 @@ rtype_from_file_extension(const char *fn)
 /*
  * Validate that a filename listed on a Manifest only contains characters
  * permitted in draft-ietf-sidrops-6486bis section 4.2.2
+ * Also ensure that there is exactly one '.'.
  */
 static int
-valid_filename(const char *fn, size_t len)
+valid_mft_filename(const char *fn, size_t len)
 {
 	const unsigned char *c;
-	size_t i;
 
-	for (c = fn, i = 0; i < len; i++, c++)
-		if (!isalnum(*c) && *c != '-' && *c != '_' && *c != '.')
-			return 0;
+	if (!valid_filename(fn, len))
+		return 0;
 
 	c = memchr(fn, '.', len);
 	if (c == NULL || c != memrchr(fn, '.', len))
@@ -145,7 +193,7 @@ valid_filename(const char *fn, size_t len)
 }
 
 /*
- * Check that the file is a CER, CRL, GBR or a ROA.
+ * Check that the file is an ASPA, CER, CRL, GBR or a ROA.
  * Returns corresponding rtype or RTYPE_INVALID on error.
  */
 static enum rtype
@@ -155,11 +203,11 @@ rtype_from_mftfile(const char *fn)
 
 	type = rtype_from_file_extension(fn);
 	switch (type) {
+	case RTYPE_ASPA:
 	case RTYPE_CER:
 	case RTYPE_CRL:
 	case RTYPE_GBR:
 	case RTYPE_ROA:
-	case RTYPE_ASPA:
 		return type;
 	default:
 		return RTYPE_INVALID;
@@ -171,121 +219,45 @@ rtype_from_mftfile(const char *fn)
  * Return zero on failure, non-zero on success.
  */
 static int
-mft_parse_filehash(struct parse *p, const ASN1_OCTET_STRING *os)
+mft_parse_filehash(struct parse *p, const FileAndHash *fh)
 {
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*file, *hash;
 	char			*fn = NULL;
-	const unsigned char	*d = os->data;
-	size_t			 dsz = os->length;
 	int			 rc = 0;
 	struct mftfile		*fent;
+	enum rtype		 type;
 
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6486 section 4.2.1: FileAndHash: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	} else if (sk_ASN1_TYPE_num(seq) != 2) {
-		warnx("%s: RFC 6486 section 4.2.1: FileAndHash: "
-		    "want 2 elements, have %d", p->fn,
-		    sk_ASN1_TYPE_num(seq));
-		goto out;
-	}
-
-	/* First is the filename itself. */
-
-	file = sk_ASN1_TYPE_value(seq, 0);
-	if (file->type != V_ASN1_IA5STRING) {
-		warnx("%s: RFC 6486 section 4.2.1: FileAndHash: "
-		    "want ASN.1 IA5 string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(file->type), file->type);
-		goto out;
-	}
-	if (!valid_filename(file->value.ia5string->data,
-	    file->value.ia5string->length)) {
+	if (!valid_mft_filename(fh->file->data, fh->file->length)) {
 		warnx("%s: RFC 6486 section 4.2.2: bad filename", p->fn);
 		goto out;
 	}
-	fn = strndup((const char *)file->value.ia5string->data,
-	    file->value.ia5string->length);
+	fn = strndup(fh->file->data, fh->file->length);
 	if (fn == NULL)
 		err(1, NULL);
 
-	/* Now hash value. */
-
-	hash = sk_ASN1_TYPE_value(seq, 1);
-	if (hash->type != V_ASN1_BIT_STRING) {
-		warnx("%s: RFC 6486 section 4.2.1: FileAndHash: "
-		    "want ASN.1 bit string, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(hash->type), hash->type);
+	if (fh->hash->length != SHA256_DIGEST_LENGTH) {
+		warnx("%s: RFC 6486 section 4.2.1: hash: "
+		    "invalid SHA256 length, have %d",
+		    p->fn, fh->hash->length);
 		goto out;
 	}
 
-	if (hash->value.bit_string->length != SHA256_DIGEST_LENGTH) {
-		warnx("%s: RFC 6486 section 4.2.1: hash: "
-		    "invalid SHA256 length, have %d",
-		    p->fn, hash->value.bit_string->length);
-		goto out;
+	type = rtype_from_mftfile(fn);
+	/* remember the filehash for the CRL in struct mft */
+	if (type == RTYPE_CRL && strcmp(fn, p->res->crl) == 0) {
+		memcpy(p->res->crlhash, fh->hash->data, SHA256_DIGEST_LENGTH);
+		p->found_crl = 1;
 	}
 
 	/* Insert the filename and hash value. */
 	fent = &p->res->files[p->res->filesz++];
-
-	fent->type = rtype_from_mftfile(fn);
+	fent->type = type;
 	fent->file = fn;
 	fn = NULL;
-	memcpy(fent->hash, hash->value.bit_string->data, SHA256_DIGEST_LENGTH);
-
-	rc = 1;
-out:
-	free(fn);
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
-	return rc;
-}
-
-/*
- * Parse the "FileAndHash" sequence, RFC 6486, sec. 4.2.
- * Return zero on failure, non-zero on success.
- */
-static int
-mft_parse_flist(struct parse *p, const ASN1_OCTET_STRING *os)
-{
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*t;
-	const unsigned char	*d = os->data;
-	size_t			 dsz = os->length;
-	int			 i, rc = 0;
-
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6486 section 4.2: fileList: "
-		    "failed ASN.1 sequence parse", p->fn);
-		goto out;
-	}
-
-	if (sk_ASN1_TYPE_num(seq) > MAX_MANIFEST_ENTRIES) {
-		warnx("%s: %d exceeds manifest entry limit (%d)", p->fn,
-		    sk_ASN1_TYPE_num(seq), MAX_MANIFEST_ENTRIES);
-		goto out;
-	}
-
-	p->res->files = calloc(sk_ASN1_TYPE_num(seq), sizeof(struct mftfile));
-	if (p->res->files == NULL)
-		err(1, NULL);
-
-	for (i = 0; i < sk_ASN1_TYPE_num(seq); i++) {
-		t = sk_ASN1_TYPE_value(seq, i);
-		if (t->type != V_ASN1_SEQUENCE) {
-			warnx("%s: RFC 6486 section 4.2: fileList: "
-			    "want ASN.1 sequence, have %s (NID %d)",
-			    p->fn, ASN1_tag2str(t->type), t->type);
-			goto out;
-		} else if (!mft_parse_filehash(p, t->value.octet_string))
-			goto out;
-	}
+	memcpy(fent->hash, fh->hash->data, SHA256_DIGEST_LENGTH);
 
 	rc = 1;
  out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
+	free(fn);
 	return rc;
 }
 
@@ -296,58 +268,20 @@ mft_parse_flist(struct parse *p, const ASN1_OCTET_STRING *os)
 static int
 mft_parse_econtent(const unsigned char *d, size_t dsz, struct parse *p)
 {
-	ASN1_SEQUENCE_ANY	*seq;
-	const ASN1_TYPE		*t;
-	const ASN1_GENERALIZEDTIME *from, *until;
-	long			 mft_version;
-	int			 i = 0, rc = 0;
+	Manifest		*mft;
+	FileAndHash		*fh;
+	int			 i, rc = 0;
 
-	if ((seq = d2i_ASN1_SEQUENCE_ANY(NULL, &d, dsz)) == NULL) {
-		cryptowarnx("%s: RFC 6486 section 4.2: Manifest: "
-		    "failed ASN.1 sequence parse", p->fn);
+	if ((mft = d2i_Manifest(NULL, &d, dsz)) == NULL) {
+		cryptowarnx("%s: RFC 6486 section 4: failed to parse Manifest",
+		    p->fn);
 		goto out;
 	}
 
-	/* Test if the optional profile version field is present. */
-	if (sk_ASN1_TYPE_num(seq) != 5 &&
-	    sk_ASN1_TYPE_num(seq) != 6) {
-		warnx("%s: RFC 6486 section 4.2: Manifest: "
-		    "want 5 or 6 elements, have %d", p->fn,
-		    sk_ASN1_TYPE_num(seq));
+	if (!valid_econtent_version(p->fn, mft->version))
 		goto out;
-	}
 
-	/* Parse the optional version field */
-	if (sk_ASN1_TYPE_num(seq) == 6) {
-		t = sk_ASN1_TYPE_value(seq, i++);
-		d = t->value.asn1_string->data;
-		dsz = t->value.asn1_string->length;
-
-		if (cms_econtent_version(p->fn, &d, dsz, &mft_version) == -1)
-			goto out;
-
-		switch (mft_version) {
-		case 0:
-			warnx("%s: incorrect encoding for version 0", p->fn);
-			goto out;
-		default:
-			warnx("%s: version %ld not supported (yet)", p->fn,
-			    mft_version);
-			goto out;
-		}
-	}
-
-	/* Now the manifest sequence number. */
-
-	t = sk_ASN1_TYPE_value(seq, i++);
-	if (t->type != V_ASN1_INTEGER) {
-		warnx("%s: RFC 6486 section 4.2.1: manifestNumber: "
-		    "want ASN.1 integer, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-
-	p->res->seqnum = x509_convert_seqnum(p->fn, t->value.integer);
+	p->res->seqnum = x509_convert_seqnum(p->fn, mft->manifestNumber);
 	if (p->res->seqnum == NULL)
 		goto out;
 
@@ -361,60 +295,42 @@ mft_parse_econtent(const unsigned char *d, size_t dsz, struct parse *p)
 	 * compare against the current time trivially.
 	 */
 
-	t = sk_ASN1_TYPE_value(seq, i++);
-	if (t->type != V_ASN1_GENERALIZEDTIME) {
-		warnx("%s: RFC 6486 section 4.2.1: thisUpdate: "
-		    "want ASN.1 generalised time, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	from = t->value.generalizedtime;
-
-	t = sk_ASN1_TYPE_value(seq, i++);
-	if (t->type != V_ASN1_GENERALIZEDTIME) {
-		warnx("%s: RFC 6486 section 4.2.1: nextUpdate: "
-		    "want ASN.1 generalised time, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	until = t->value.generalizedtime;
-
-	if (!mft_parse_time(from, until, p))
+	if (!mft_parse_time(mft->thisUpdate, mft->nextUpdate, p))
 		goto out;
 
-	/* File list algorithm. */
-
-	t = sk_ASN1_TYPE_value(seq, i++);
-	if (t->type != V_ASN1_OBJECT) {
-		warnx("%s: RFC 6486 section 4.2.1: fileHashAlg: "
-		    "want ASN.1 object, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
-		goto out;
-	}
-	if (OBJ_obj2nid(t->value.object) != NID_sha256) {
+	if (OBJ_obj2nid(mft->fileHashAlg) != NID_sha256) {
 		warnx("%s: RFC 6486 section 4.2.1: fileHashAlg: "
 		    "want SHA256 object, have %s (NID %d)", p->fn,
-		    ASN1_tag2str(OBJ_obj2nid(t->value.object)),
-		    OBJ_obj2nid(t->value.object));
+		    ASN1_tag2str(OBJ_obj2nid(mft->fileHashAlg)),
+		    OBJ_obj2nid(mft->fileHashAlg));
 		goto out;
 	}
 
-	/* Now the sequence. */
-
-	t = sk_ASN1_TYPE_value(seq, i++);
-	if (t->type != V_ASN1_SEQUENCE) {
-		warnx("%s: RFC 6486 section 4.2.1: fileList: "
-		    "want ASN.1 sequence, have %s (NID %d)",
-		    p->fn, ASN1_tag2str(t->type), t->type);
+	if (sk_FileAndHash_num(mft->fileList) >= MAX_MANIFEST_ENTRIES) {
+		warnx("%s: %d exceeds manifest entry limit (%d)", p->fn,
+		    sk_FileAndHash_num(mft->fileList), MAX_MANIFEST_ENTRIES);
 		goto out;
 	}
 
-	if (!mft_parse_flist(p, t->value.octet_string))
+	p->res->files = calloc(sk_FileAndHash_num(mft->fileList),
+	    sizeof(struct mftfile));
+	if (p->res->files == NULL)
+		err(1, NULL);
+
+	for (i = 0; i < sk_FileAndHash_num(mft->fileList); i++) {
+		fh = sk_FileAndHash_value(mft->fileList, i);
+		if (!mft_parse_filehash(p, fh))
+			goto out;
+	}
+
+	if (!p->found_crl) {
+		warnx("%s: CRL not part of MFT fileList", p->fn);
 		goto out;
+	}
 
 	rc = 1;
-out:
-	sk_ASN1_TYPE_pop_free(seq, ASN1_TYPE_free);
+ out:
+	Manifest_free(mft);
 	return rc;
 }
 
@@ -432,6 +348,7 @@ mft_parse(X509 **x509, const char *fn, const unsigned char *der, size_t len)
 	int		 rc = 0;
 	size_t		 cmsz;
 	unsigned char	*cms;
+	char		*crldp = NULL, *crlfile;
 
 	memset(&p, 0, sizeof(struct parse));
 	p.fn = fn;
@@ -456,6 +373,29 @@ mft_parse(X509 **x509, const char *fn, const unsigned char *der, size_t len)
 		goto out;
 	}
 
+	if (!x509_inherits(*x509)) {
+		warnx("%s: RFC 3779 extension not set to inherit", fn);
+		goto out;
+	}
+
+	/* get CRL info for later */
+	if (!x509_get_crl(*x509, fn, &crldp))
+		goto out;
+	if (crldp == NULL) {
+		warnx("%s: RFC 6487 section 4.8.6: CRL: "
+		    "missing CRL distribution point extension", fn);
+		goto out;
+	}
+	if ((crlfile = strrchr(crldp, '/')) == NULL ||
+	    !valid_mft_filename(crlfile + 1, strlen(crlfile + 1)) ||
+	    rtype_from_file_extension(crlfile + 1) != RTYPE_CRL) {
+		warnx("%s: RFC 6487 section 4.8.6: CRL: "
+		    "bad CRL distribution point extension", fn);
+		goto out;
+	}
+	if ((p.res->crl = strdup(crlfile + 1)) == NULL)
+		err(1, NULL);
+
 	if (mft_parse_econtent(cms, cmsz, &p) == 0)
 		goto out;
 
@@ -467,6 +407,7 @@ out:
 		X509_free(*x509);
 		*x509 = NULL;
 	}
+	free(crldp);
 	free(cms);
 	return p.res;
 }
