@@ -1,4 +1,4 @@
-/*	$OpenBSD: output-json.c,v 1.23 2022/01/14 15:00:23 claudio Exp $ */
+/*	$OpenBSD: output-json.c,v 1.26 2022/05/15 16:43:34 tb Exp $ */
 /*
  * Copyright (c) 2019 Claudio Jeker <claudio@openbsd.org>
  *
@@ -25,14 +25,13 @@
 static int
 outputheader_json(FILE *out, struct stats *st)
 {
-	char		hn[NI_MAXHOST], tbuf[26];
+	char		 hn[NI_MAXHOST], tbuf[26];
 	struct tm	*tp;
-	time_t		t;
-	size_t		i;
+	time_t		 t;
+	int		 i;
 
 	time(&t);
-	setenv("TZ", "UTC", 1);
-	tp = localtime(&t);
+	tp = gmtime(&t);
 	strftime(tbuf, sizeof tbuf, "%FT%TZ", tp);
 
 	gethostname(hn, sizeof hn);
