@@ -1,4 +1,4 @@
-/*	$OpenBSD: rsync.c,v 1.32 2022/01/13 11:50:29 claudio Exp $ */
+/*	$OpenBSD: rsync.c,v 1.38 2022/05/24 09:20:49 claudio Exp $ */
 /*
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -44,7 +44,7 @@
 struct	rsyncproc {
 	char		*uri; /* uri of this rsync proc */
 	unsigned int	 id; /* identity of request */
-	pid_t	 	 pid; /* pid of process or 0 if unassociated */
+	pid_t		 pid; /* pid of process or 0 if unassociated */
 };
 
 /*
@@ -137,23 +137,22 @@ proc_child(int signal)
  * does so.
  * It then responds with the identifier of the repo that it updated.
  * It only exits cleanly when fd is closed.
- * FIXME: limit the number of simultaneous process.
- * Currently, an attacker can trivially specify thousands of different
- * repositories and saturate our system.
  */
 void
 proc_rsync(char *prog, char *bind_addr, int fd)
 {
-	size_t			 i, idsz = 0, nprocs = 0;
+	size_t			 i, nprocs = 0;
 	int			 rc = 0;
 	struct pollfd		 pfd;
 	struct msgbuf		 msgq;
 	struct ibuf		*b, *inbuf = NULL;
 	sigset_t		 mask, oldmask;
-	struct rsyncproc	*ids = NULL;
+	struct rsyncproc	 ids[MAX_RSYNC_REQUESTS] = { 0 };
+
+	if (pledge("stdio rpath proc exec unveil", NULL) == -1)
+		err(1, "pledge");
 
 	pfd.fd = fd;
-
 	msgbuf_init(&msgq);
 	msgq.fd = fd;
 
@@ -212,7 +211,7 @@ proc_rsync(char *prog, char *bind_addr, int fd)
 		int st;
 
 		pfd.events = 0;
-		if (nprocs < MAX_RSYNC_PROCESSES)
+		if (nprocs < MAX_RSYNC_REQUESTS)
 			pfd.events |= POLLIN;
 		if (msgq.queued)
 			pfd.events |= POLLOUT;
@@ -231,10 +230,10 @@ proc_rsync(char *prog, char *bind_addr, int fd)
 			while ((pid = waitpid(WAIT_ANY, &st, WNOHANG)) > 0) {
 				int ok = 1;
 
-				for (i = 0; i < idsz; i++)
+				for (i = 0; i < MAX_RSYNC_REQUESTS; i++)
 					if (ids[i].pid == pid)
 						break;
-				if (i >= idsz)
+				if (i >= MAX_RSYNC_REQUESTS)
 					errx(1, "waitpid: %d unexpected", pid);
 
 				if (!WIFEXITED(st)) {
@@ -249,7 +248,7 @@ proc_rsync(char *prog, char *bind_addr, int fd)
 
 				b = io_new_buffer();
 				io_simple_buffer(b, &ids[i].id,
-				     sizeof(ids[i].id));
+				    sizeof(ids[i].id));
 				io_simple_buffer(b, &ok, sizeof(ok));
 				io_close_buffer(&msgq, b);
 
@@ -279,6 +278,8 @@ proc_rsync(char *prog, char *bind_addr, int fd)
 
 		if (!(pfd.revents & POLLIN))
 			continue;
+		if (nprocs >= MAX_RSYNC_REQUESTS)
+			continue;
 
 		b = io_buf_read(fd, &inbuf);
 		if (b == NULL)
@@ -318,6 +319,7 @@ proc_rsync(char *prog, char *bind_addr, int fd)
 			args[i++] = "--include=*.gbr";
 			args[i++] = "--include=*.mft";
 			args[i++] = "--include=*.roa";
+			args[i++] = "--include=*.asa";
 			args[i++] = "--exclude=*";
 			if (bind_addr != NULL) {
 				args[i++] = "--address";
@@ -338,16 +340,10 @@ proc_rsync(char *prog, char *bind_addr, int fd)
 
 		/* Augment the list of running processes. */
 
-		for (i = 0; i < idsz; i++)
+		for (i = 0; i < MAX_RSYNC_REQUESTS; i++)
 			if (ids[i].pid == 0)
 				break;
-		if (i == idsz) {
-			ids = reallocarray(ids, idsz + 1, sizeof(*ids));
-			if (ids == NULL)
-				err(1, NULL);
-			idsz++;
-		}
-
+		assert(i < MAX_RSYNC_REQUESTS);
 		ids[i].id = id;
 		ids[i].pid = pid;
 		ids[i].uri = uri;
@@ -360,13 +356,12 @@ proc_rsync(char *prog, char *bind_addr, int fd)
 	}
 
 	/* No need for these to be hanging around. */
-	for (i = 0; i < idsz; i++)
-		if (ids[i].pid > 0) {
+	for (i = 0; i < MAX_RSYNC_REQUESTS; i++)
+		if (ids[i].pid != 0) {
 			kill(ids[i].pid, SIGTERM);
 			free(ids[i].uri);
 		}
 
 	msgbuf_clear(&msgq);
-	free(ids);
 	exit(rc);
 }
