@@ -1,5 +1,6 @@
-/*	$OpenBSD: x509.c,v 1.38 2022/04/01 17:22:07 claudio Exp $ */
+/*	$OpenBSD: x509.c,v 1.46 2022/05/31 18:51:35 tb Exp $ */
 /*
+ * Copyright (c) 2022 Theo Buehler <tb@openbsd.org>
  * Copyright (c) 2021 Claudio Jeker <claudio@openbsd.org>
  * Copyright (c) 2019 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -42,6 +43,7 @@ ASN1_OBJECT	*cnt_type_oid;	/* pkcs-9 id-contentType */
 ASN1_OBJECT	*msg_dgst_oid;	/* pkcs-9 id-messageDigest */
 ASN1_OBJECT	*sign_time_oid;	/* pkcs-9 id-signingTime */
 ASN1_OBJECT	*bin_sign_time_oid;	/* pkcs-9 id-aa-binarySigningTime */
+ASN1_OBJECT	*rsc_oid;	/* id-ct-signedChecklist */
 
 void
 x509_init_oid(void)
@@ -76,6 +78,9 @@ x509_init_oid(void)
 	    OBJ_txt2obj("1.2.840.113549.1.9.16.2.46", 1)) == NULL)
 		errx(1, "OBJ_txt2obj for %s failed",
 		    "1.2.840.113549.1.9.16.2.46");
+	if ((rsc_oid = OBJ_txt2obj("1.2.840.113549.1.9.16.1.48", 1)) == NULL)
+		errx(1, "OBJ_txt2obj for %s failed",
+		    "1.2.840.113549.1.9.16.1.48");
 }
 
 /*
@@ -179,13 +184,18 @@ x509_get_purpose(X509 *x, const char *fn)
 {
 	EXTENDED_KEY_USAGE		*eku = NULL;
 	int				 crit;
-	enum cert_purpose		 purpose = 0;
+	enum cert_purpose		 purpose = CERT_PURPOSE_INVALID;
 
 	if (X509_check_ca(x) == 1) {
 		purpose = CERT_PURPOSE_CA;
 		goto out;
 	}
 
+	if (X509_get_extension_flags(x) & EXFLAG_BCONS) {
+		warnx("%s: Basic Constraints ext in non-CA cert", fn);
+		goto out;
+	}
+
 	eku = X509_get_ext_d2i(x, NID_ext_key_usage, &crit, NULL);
 	if (eku == NULL) {
 		warnx("%s: EKU: extension missing", fn);
@@ -306,24 +316,10 @@ x509_get_aia(X509 *x, const char *fn, char **aia)
 		    "expected caIssuers, have %d", fn, OBJ_obj2nid(ad->method));
 		goto out;
 	}
-	if (ad->location->type != GEN_URI) {
-		warnx("%s: RFC 6487 section 4.8.7: AIA: "
-		    "want GEN_URI type, have %d", fn, ad->location->type);
-		goto out;
-	}
 
-	if (ASN1_STRING_length(ad->location->d.uniformResourceIdentifier)
-	    > MAX_URI_LENGTH) {
-		warnx("%s: RFC 6487 section 4.8.7: AIA: "
-		    "URI exceeds max length of %d", fn, MAX_URI_LENGTH);
+	if (!x509_location(fn, "AIA: caIssuers", NULL, ad->location, aia))
 		goto out;
-	}
 
-	*aia = strndup(
-	    ASN1_STRING_get0_data(ad->location->d.uniformResourceIdentifier),
-	    ASN1_STRING_length(ad->location->d.uniformResourceIdentifier));
-	if (*aia == NULL)
-		err(1, NULL);
 	rc = 1;
 
 out:
@@ -349,7 +345,51 @@ x509_get_expire(X509 *x, const char *fn, time_t *tt)
 		return 0;
 	}
 	return 1;
+}
 
+/*
+ * Check whether the RFC 3779 extensions are set to inherit.
+ * Return 1 if both AS & IP are set to inherit.
+ * Return 0 on failure (such as missing extensions or no inheritance).
+ */
+int
+x509_inherits(X509 *x)
+{
+	STACK_OF(IPAddressFamily)	*addrblk = NULL;
+	ASIdentifiers			*asidentifiers = NULL;
+	const IPAddressFamily		*af;
+	int				 i, rc = 0;
+
+	addrblk = X509_get_ext_d2i(x, NID_sbgp_ipAddrBlock, NULL, NULL);
+	if (addrblk == NULL)
+		goto out;
+
+	/*
+	 * Check by hand, since X509v3_addr_inherits() success only means that
+	 * at least one address family inherits, not all of them.
+	 */
+	for (i = 0; i < sk_IPAddressFamily_num(addrblk); i++) {
+		af = sk_IPAddressFamily_value(addrblk, i);
+		if (af->ipAddressChoice->type != IPAddressChoice_inherit)
+			goto out;
+	}
+
+	asidentifiers = X509_get_ext_d2i(x, NID_sbgp_autonomousSysNum, NULL,
+	    NULL);
+	if (asidentifiers == NULL)
+		goto out;
+
+	/* We need to have AS numbers and don't want RDIs. */
+	if (asidentifiers->asnum == NULL || asidentifiers->rdi != NULL)
+		goto out;
+	if (!X509v3_asid_inherits(asidentifiers))
+		goto out;
+
+	rc = 1;
+ out:
+	ASIdentifiers_free(asidentifiers);
+	sk_IPAddressFamily_pop_free(addrblk, IPAddressFamily_free);
+	return rc;
 }
 
 /*
@@ -364,8 +404,9 @@ x509_get_crl(X509 *x, const char *fn, char **crl)
 {
 	CRL_DIST_POINTS		*crldp;
 	DIST_POINT		*dp;
+	GENERAL_NAMES		*names;
 	GENERAL_NAME		*name;
-	int			 crit, rc = 0;
+	int			 i, crit, rc = 0;
 
 	*crl = NULL;
 	crldp = X509_get_ext_d2i(x, NID_crl_distribution_points, &crit, NULL);
@@ -397,34 +438,25 @@ x509_get_crl(X509 *x, const char *fn, char **crl)
 		goto out;
 	}
 
-	if (sk_GENERAL_NAME_num(dp->distpoint->name.fullname) != 1) {
-		warnx("%s: RFC 6487 section 4.8.6: CRL: "
-		    "want 1 full name, have %d", fn,
-		    sk_GENERAL_NAME_num(dp->distpoint->name.fullname));
-		goto out;
+	names = dp->distpoint->name.fullname;
+	for (i = 0; i < sk_GENERAL_NAME_num(names); i++) {
+		name = sk_GENERAL_NAME_value(names, i);
+		/* Don't warn on non-rsync URI, so check this afterward. */
+		if (!x509_location(fn, "CRL distribution point", NULL, name,
+		    crl))
+			goto out;
+		if (strncasecmp(*crl, "rsync://", 8) == 0) {
+			rc = 1;
+			goto out;
+		}
+		free(*crl);
+		*crl = NULL;
 	}
 
-	name = sk_GENERAL_NAME_value(dp->distpoint->name.fullname, 0);
-	if (name->type != GEN_URI) {
-		warnx("%s: RFC 6487 section 4.8.6: CRL: "
-		    "want URI type, have %d", fn, name->type);
-		goto out;
-	}
+	warnx("%s: RFC 6487 section 4.8.6: no rsync URI "
+	    "in CRL distributionPoint", fn);
 
-	if (ASN1_STRING_length(name->d.uniformResourceIdentifier)
-	    > MAX_URI_LENGTH) {
-		warnx("%s: RFC 6487 section 4.8.6: CRL: "
-		    "URI exceeds max length of %d", fn, MAX_URI_LENGTH);
-		goto out;
-	}
-
-	*crl = strndup(ASN1_STRING_get0_data(name->d.uniformResourceIdentifier),
-	    ASN1_STRING_length(name->d.uniformResourceIdentifier));
-	if (*crl == NULL)
-		err(1, NULL);
-	rc = 1;
-
-out:
+ out:
 	CRL_DIST_POINTS_free(crldp);
 	return rc;
 }
@@ -498,8 +530,42 @@ x509_get_time(const ASN1_TIME *at, time_t *t)
 	memset(&tm, 0, sizeof(tm));
 	if (ASN1_time_parse(at->data, at->length, &tm, 0) == -1)
 		return 0;
-	if ((*t = mktime(&tm)) == -1)
-		errx(1, "mktime failed");
+	if ((*t = timegm(&tm)) == -1)
+		errx(1, "timegm failed");
+	return 1;
+}
+
+/*
+ * Extract and validate an accessLocation, RFC 6487, 4.8 and RFC 8192, 3.2.
+ * Returns 0 on failure and 1 on success.
+ */
+int
+x509_location(const char *fn, const char *descr, const char *proto,
+    GENERAL_NAME *location, char **out)
+{
+	ASN1_IA5STRING	*uri;
+
+	if (*out != NULL) {
+		warnx("%s: RFC 6487 section 4.8: %s already specified", fn,
+		    descr);
+		return 0;
+	}
+
+	if (location->type != GEN_URI) {
+		warnx("%s: RFC 6487 section 4.8: %s not URI", fn, descr);
+		return 0;
+	}
+
+	uri = location->d.uniformResourceIdentifier;
+
+	if (!valid_uri(uri->data, uri->length, proto)) {
+		warnx("%s: RFC 6487 section 4.8: %s bad location", fn, descr);
+		return 0;
+	}
+
+	if ((*out = strndup(uri->data, uri->length)) == NULL)
+		err(1, NULL);
+
 	return 1;
 }
 
